import unittest

import numpy as np
from nose.plugins.attrib import attr

import yt
from yt.testing import ANSWER_TEST_TAG, fake_amr_ds, requires_module
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.geo_plot_utils import get_mpl_transform, transform_list


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def compare(
    ds, field, idir, test_prefix, test_name, projection, decimals=12, annotate=False
):
    def slice_image(filename_prefix):
        sl = yt.SlicePlot(ds, idir, field, origin="native")
        sl.set_mpl_projection(projection)
        if annotate:
            sl._setup_plots()
            sl.annotate_mesh_lines()
        sl.set_log("all", False)
        image_file = sl.save(filename_prefix)
        return image_file

    slice_image.__name__ = f"slice_{test_prefix}"
    test = GenericImageTest(ds, slice_image, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test


@requires_module("cartopy")
@attr(ANSWER_TEST_TAG)
def test_geo_slices_amr():
    ds = fake_amr_ds(geometry="geographic")
    for transform in transform_list:
        if transform == "UTM":
            # requires additional argument so we skip
            continue
        if transform == "OSNI":
            # avoid crashes, see https://github.com/SciTools/cartopy/issues/1177
            continue
        for field in ds.field_list:
            prefix = f"{field[0]}_{field[1]}_{transform}"
            yield compare(
                ds,
                field,
                "altitude",
                test_prefix=prefix,
                test_name="geo_slices_amr",
                projection=transform,
            )


class TestGeoProjections(unittest.TestCase):
    @requires_module("cartopy")
    def setUp(self):
        self.ds = fake_amr_ds(geometry="geographic")
        # switch off the log plot to avoid some unrelated matplotlib issues
        f = self.ds._get_field_info("stream", "Density")
        f.take_log = False

    @requires_module("cartopy")
    def tearDown(self):
        del self.ds

    @requires_module("cartopy")
    def test_geo_projection_setup(self):

        from yt.utilities.on_demand_imports import _cartopy as cartopy

        axis = "altitude"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        assert isinstance(self.slc._projection, cartopy.crs.Mollweide)
        assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
        assert self.ds.coordinates.data_projection[axis] == "Mollweide"
        assert self.ds.coordinates.data_transform[axis] == "PlateCarree"
        assert isinstance(
            self.slc._projection,
            type(self.slc.plots[("stream", "Density")].axes.projection),
        )

    @requires_module("cartopy")
    def test_geo_projections(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        self.slc = yt.SlicePlot(
            self.ds, "altitude", ("stream", "Density"), origin="native"
        )

        for transform in transform_list:
            if transform == "UTM":
                # this requires special arguments so let's skip it
                continue
            if transform == "OSNI":
                # avoid crashes, see https://github.com/SciTools/cartopy/issues/1177
                continue
            self.slc.set_mpl_projection(transform)
            proj_type = type(get_mpl_transform(transform))

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_projection_object(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        shortlist = ["Orthographic", "PlateCarree", "Mollweide"]

        for transform in shortlist:
            projection = get_mpl_transform(transform)
            proj_type = type(projection)
            self.slc = yt.SlicePlot(
                self.ds, "altitude", ("stream", "Density"), origin="native"
            )
            self.slc.set_mpl_projection(projection)

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_nondefault_transform(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        axis = "altitude"
        # Note: The Miller transform has an extent of approx. +/- 180 in x,
        # +/-132 in y (in Miller, x is longitude, y is a factor of latitude).
        # So by changing the projection in this way, the dataset goes from
        # covering the whole globe (+/- 90 latitude), to covering part of the
        # globe (+/-72 latitude). Totally fine for testing that the code runs,
        # but good to be aware of!
        self.ds.coordinates.data_transform[axis] = "Miller"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        shortlist = ["Orthographic", "PlateCarree", "Mollweide"]

        for transform in shortlist:

            self.slc.set_mpl_projection(transform)
            proj_type = type(get_mpl_transform(transform))

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.Miller)
            assert self.ds.coordinates.data_projection[axis] == "Mollweide"
            assert self.ds.coordinates.data_transform[axis] == "Miller"
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_extent(self):
        # checks that the axis extent is narrowed when doing a subselection
        axis = "altitude"
        slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")
        ax = slc.plots[("stream", "Density")].axes
        full_extent = np.abs(ax.get_extent())

        slc = yt.SlicePlot(
            self.ds, axis, ("stream", "Density"), origin="native", width=(80.0, 50.0)
        )
        ax = slc.plots[("stream", "Density")].axes
        extent = np.abs(ax.get_extent())
        assert np.all(extent < full_extent)


class TestNonGeoProjections(unittest.TestCase):
    def setUp(self):
        self.ds = fake_amr_ds()

    def tearDown(self):
        del self.ds
        del self.slc

    def test_projection_setup(self):
        axis = "x"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        assert self.ds.coordinates.data_projection[axis] is None
        assert self.ds.coordinates.data_transform[axis] is None
        assert self.slc._projection is None
        assert self.slc._transform is None



import os
import shutil
import tempfile
import unittest

import matplotlib.pyplot as plt
import numpy as np
from nose.tools import assert_raises

from yt import make_colormap, show_colormaps
from yt.testing import assert_almost_equal, assert_equal, requires_backend


class TestColorMaps(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

    @requires_backend("Agg")
    def test_show_colormaps(self):
        show_colormaps()
        show_colormaps(subset=["jet", "cool"])
        show_colormaps(subset="yt_native", filename="yt_color_maps.png")

        # Test for non-existent color map
        with assert_raises(AttributeError) as ex:
            show_colormaps(subset="unknown", filename="yt_color_maps.png")
        desired = (
            "show_colormaps requires subset attribute to be 'all', "
            "'yt_native', or a list of valid colormap names."
        )
        assert_equal(str(ex.exception), desired)

    @requires_backend("Agg")
    def test_make_colormap(self):
        make_colormap(
            [([0, 0, 1], 10), ([1, 1, 1], 10), ([1, 0, 0], 10)],
            name="french_flag",
            interpolate=False,
        )
        show_colormaps("french_flag")

        cmap = make_colormap(
            [("dred", 5), ("blue", 2.0), ("orange", 0)], name="my_cmap"
        )
        assert_almost_equal(
            cmap["red"][1], np.array([0.00392157, 0.62400345, 0.62400345])
        )

        assert_almost_equal(
            cmap["blue"][2], np.array([0.00784314, 0.01098901, 0.01098901])
        )

        assert_almost_equal(cmap["green"][3], np.array([0.01176471, 0.0, 0.0]))


def test_cmyt_integration():
    for name in ["algae", "bds_highcontrast", "kelp", "arbre", "octarine", "kamae"]:
        cmap = plt.get_cmap(name)
        assert cmap.name == name
        name_r = name + "_r"
        cmap_r = plt.get_cmap(name_r)
        assert cmap_r.name == name_r

    for name in ["algae", "kelp", "arbre", "octarine", "pastel"]:
        cmap = plt.get_cmap("cmyt." + name)
        assert cmap.name == "cmyt." + name



import yt

# This shows how to save ImageArray objects, such as those returned from
# volume renderings, to pngs with varying backgrounds.

# First we use the simple_volume_rendering.py recipe from above to generate
# a standard volume rendering.

ds = yt.load("Enzo_64/DD0043/data0043")
im, sc = yt.volume_render(ds, ("gas", "density"))
im.write_png("original.png", sigma_clip=8.0)

# Our image array can now be transformed to include different background
# colors.  By default, the background color is black.  The following
# modifications can be used on any image array.

# write_png accepts a background keyword argument that defaults to 'black'.
# Other choices include:
# black (0.,0.,0.,1.)
# white (1.,1.,1.,1.)
# None  (0.,0.,0.,0.) <-- Transparent!
# any rgba list/array: [r,g,b,a], bounded by 0..1

# We include the sigma_clip=8 keyword here to bring out more contrast between
# the background and foreground, but it is entirely optional.

im.write_png("black_bg.png", background="black", sigma_clip=8.0)
im.write_png("white_bg.png", background="white", sigma_clip=8.0)
im.write_png("green_bg.png", background=[0.0, 1.0, 0.0, 1.0], sigma_clip=8.0)
im.write_png("transparent_bg.png", background=None, sigma_clip=8.0)



import os
import os.path
import shutil
import tempfile

import numpy as np

import yt
from yt.testing import assert_equal
from yt.utilities.lib.api import add_rgba_points_to_image  # type: ignore
from yt.visualization.color_maps import _get_cmap


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_splat():
    # Perform I/O in safe place instead of yt main dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    prng = np.random.RandomState(0x4D3D3D3)
    N = 16
    Np = int(1e2)
    image = np.zeros([N, N, 4])
    xs = prng.random_sample(Np)
    ys = prng.random_sample(Np)

    cbx = _get_cmap("RdBu")
    cs = cbx(prng.random_sample(Np))
    add_rgba_points_to_image(image, xs, ys, cs)

    before_hash = image.copy()
    fn = "tmp.png"
    yt.write_bitmap(image, fn)
    assert_equal(os.path.exists(fn), True)
    os.remove(fn)
    assert_equal(before_hash, image)

    os.chdir(curdir)
    # clean up
    shutil.rmtree(tmpdir)



import os
import shutil
import tempfile
import unittest

from nose.plugins.attrib import attr

import yt
from yt.data_objects.profiles import create_profile
from yt.testing import (
    ANSWER_TEST_TAG,
    assert_allclose_units,
    assert_array_almost_equal,
    fake_random_ds,
)
from yt.utilities.answer_testing.framework import (
    GenericImageTest,
    PhasePlotAttributeTest,
)
from yt.visualization.profile_plotter import PhasePlot, ProfilePlot

ATTR_ARGS = {
    "annotate_text": [
        (((5e-29, 5e7), "Hello YT"), {}),
        (((5e-29, 5e7), "Hello YT"), {"color": "b"}),
    ],
    "set_title": [((("gas", "mass"), "A phase plot."), {})],
    "set_log": [((("gas", "mass"), False), {})],
    "set_unit": [((("gas", "mass"), "Msun"), {})],
    "set_xlim": [((1e-27, 1e-24), {})],
    "set_ylim": [((1e2, 1e6), {})],
}


def compare(ds, plot, test_prefix, test_name, decimals=12):
    def image_from_plot(filename_prefix):
        return plot.save(filename_prefix)

    image_from_plot.__name__ = f"profile_{test_prefix}"
    test = GenericImageTest(ds, image_from_plot, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test


@attr(ANSWER_TEST_TAG)
def test_phase_plot_attributes():
    """

    This iterates over the all the plot modification functions in
    ATTR_ARGS. Each time, it compares the images produced by
    PhasePlot to the gold standard.


    """

    x_field = ("gas", "density")
    y_field = ("gas", "temperature")
    z_field = ("gas", "mass")
    decimals = 12
    ds = fake_random_ds(16, fields=("density", "temperature"), units=("g/cm**3", "K"))
    for attr_name in ATTR_ARGS.keys():
        for args in ATTR_ARGS[attr_name]:
            test = PhasePlotAttributeTest(
                ds, x_field, y_field, z_field, attr_name, args, decimals
            )
            test.prefix = f"{attr_name}_{args}"
            test.answer_name = "phase_plot_attributes"
            yield test


@attr(ANSWER_TEST_TAG)
def test_profile_plot():
    fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
    units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
    test_ds = fake_random_ds(16, fields=fields, units=units)
    regions = [test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), test_ds.all_data()]
    pr_fields = [
        [("gas", "density"), ("gas", "temperature")],
        [("gas", "density"), ("gas", "velocity_x")],
        [("gas", "temperature"), ("gas", "mass")],
        [("gas", "density"), ("index", "radius")],
        [("gas", "velocity_magnitude"), ("gas", "mass")],
    ]
    profiles = []
    for reg in regions:
        for x_field, y_field in pr_fields:
            profiles.append(ProfilePlot(reg, x_field, y_field))
            profiles.append(
                ProfilePlot(reg, x_field, y_field, fractional=True, accumulation=True)
            )
            p1d = create_profile(reg, x_field, y_field)
            profiles.append(ProfilePlot.from_profiles(p1d))
    p1 = create_profile(test_ds.all_data(), ("gas", "density"), ("gas", "temperature"))
    p2 = create_profile(test_ds.all_data(), ("gas", "density"), ("gas", "velocity_x"))
    profiles.append(
        ProfilePlot.from_profiles([p1, p2], labels=["temperature", "velocity"])
    )
    profiles[0]._repr_html_()
    for idx, plot in enumerate(profiles):
        test_prefix = f"{plot.plots.keys()}_{idx}"
        yield compare(test_ds, plot, test_prefix=test_prefix, test_name="profile_plots")


@attr(ANSWER_TEST_TAG)
def test_phase_plot():
    fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
    units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
    test_ds = fake_random_ds(16, fields=fields, units=units)
    regions = [test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), test_ds.all_data()]
    phases = []
    ph_fields = [
        [("gas", "density"), ("gas", "temperature"), ("gas", "mass")],
        [("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")],
        [("index", "radius"), ("gas", "temperature"), ("gas", "velocity_magnitude")],
    ]
    for reg in regions:
        for x_field, y_field, z_field in ph_fields:
            # set n_bins to [16, 16] since matplotlib's postscript
            # renderer is slow when it has to write a lot of polygons
            phases.append(
                PhasePlot(reg, x_field, y_field, z_field, x_bins=16, y_bins=16)
            )
            phases.append(
                PhasePlot(
                    reg,
                    x_field,
                    y_field,
                    z_field,
                    fractional=True,
                    accumulation=True,
                    x_bins=16,
                    y_bins=16,
                )
            )
            p2d = create_profile(reg, [x_field, y_field], z_field, n_bins=[16, 16])
            phases.append(PhasePlot.from_profile(p2d))
    pp = PhasePlot(
        test_ds.all_data(),
        ("gas", "density"),
        ("gas", "temperature"),
        ("gas", "mass"),
    )
    pp.set_xlim(0.3, 0.8)
    pp.set_ylim(0.4, 0.6)
    pp._setup_plots()
    xlim = pp.plots[("gas", "mass")].axes.get_xlim()
    ylim = pp.plots[("gas", "mass")].axes.get_ylim()
    assert_array_almost_equal(xlim, (0.3, 0.8))
    assert_array_almost_equal(ylim, (0.4, 0.6))
    phases.append(pp)
    phases[0]._repr_html_()
    for idx, plot in enumerate(phases):
        test_prefix = f"{plot.plots.keys()}_{idx}"
        yield compare(test_ds, plot, test_prefix=test_prefix, test_name="phase_plots")


@attr(ANSWER_TEST_TAG)
def test_profile_plot_multiple_field_multiple_plot():
    fields = ("density", "temperature", "dark_matter_density")
    units = ("g/cm**3", "K", "g/cm**3")
    ds = fake_random_ds(16, fields=fields, units=units)
    sphere = ds.sphere("max", (1.0, "Mpc"))
    profiles = []
    profiles.append(
        yt.create_profile(
            sphere, [("index", "radius")], fields=[("gas", "density")], n_bins=32
        )
    )
    profiles.append(
        yt.create_profile(
            sphere, [("index", "radius")], fields=[("gas", "density")], n_bins=64
        )
    )
    profiles.append(
        yt.create_profile(
            sphere,
            [("index", "radius")],
            fields=[("gas", "dark_matter_density")],
            n_bins=64,
        )
    )

    plot = yt.ProfilePlot.from_profiles(profiles)
    yield compare(
        ds,
        plot,
        test_prefix=plot.plots.keys(),
        test_name="profile_plot_multiple_field_multiple_plot",
    )


def test_set_units():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, ("index", "radius"), ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))
    # make sure we can set the units using the tuple without erroring out
    p1.set_unit(("gas", "density"), "Msun/kpc**3")
    p2.set_unit(("gas", "temperature"), "R")


def test_set_labels():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    plot = yt.ProfilePlot(
        ad,
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    # make sure we can set the labels without erroring out
    plot.set_ylabel("all", "test ylabel")
    plot.set_xlabel("test xlabel")


def test_create_from_dataset():
    ds = fake_random_ds(16)
    plot1 = yt.ProfilePlot(
        ds,
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    plot2 = yt.ProfilePlot(
        ds.all_data(),
        ("index", "radius"),
        [("gas", "velocity_x"), ("gas", "density")],
        weight_field=None,
    )
    assert_allclose_units(
        plot1.profiles[0][("gas", "density")], plot2.profiles[0][("gas", "density")]
    )
    assert_allclose_units(
        plot1.profiles[0]["velocity_x"], plot2.profiles[0]["velocity_x"]
    )

    plot1 = yt.PhasePlot(ds, ("gas", "density"), ("gas", "velocity_x"), ("gas", "mass"))
    plot2 = yt.PhasePlot(
        ds.all_data(), ("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")
    )
    assert_allclose_units(plot1.profile["mass"], plot2.profile["mass"])


class TestAnnotations(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

        ds = fake_random_ds(16)
        ad = ds.all_data()
        cls.fields = [
            ("gas", "velocity_x"),
            ("gas", "velocity_y"),
            ("gas", "velocity_z"),
        ]
        cls.plot = yt.ProfilePlot(
            ad, ("index", "radius"), cls.fields, weight_field=None
        )

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_annotations(self):
        # make sure we can annotate without erroring out
        # annotate the plot with only velocity_x
        self.plot.annotate_title("velocity_x plot", self.fields[0])
        self.plot.annotate_text(1e-1, 1e1, "Annotated velocity_x")

        # annotate the plots with velocity_y and velocity_z with
        # the same annotations
        self.plot.annotate_title("Velocity Plots (Y or Z)", self.fields[1:])
        self.plot.annotate_text(1e-1, 1e1, "Annotated vel_y, vel_z", self.fields[1:])
        self.plot.save()

    def test_annotations_wrong_fields(self):
        from yt.utilities.exceptions import YTFieldNotFound

        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_title("velocity_x plot", "wrong_field_name")

        with self.assertRaises(YTFieldNotFound):
            self.plot.annotate_text(1e-1, 1e1, "Annotated text", "wrong_field_name")


def test_phaseplot_set_log():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    sp = ds.sphere("max", (1.0, "Mpc"))
    p1 = yt.ProfilePlot(sp, ("index", "radius"), ("gas", "density"))
    p2 = yt.PhasePlot(sp, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))
    # make sure we can set the log-scaling using the tuple without erroring out
    p1.set_log(("gas", "density"), False)
    p2.set_log(("gas", "temperature"), False)
    assert not p1.y_log["gas", "density"]
    assert not p2.y_log

    # make sure we can set the log-scaling using a string without erroring out
    p1.set_log(("gas", "density"), True)
    p2.set_log(("gas", "temperature"), True)
    assert p1.y_log["gas", "density"]
    assert p2.y_log

    # make sure we can set the log-scaling using a field object
    p1.set_log(ds.fields.gas.density, False)
    p2.set_log(ds.fields.gas.temperature, False)
    assert not p1.y_log["gas", "density"]
    assert not p2.y_log


def test_phaseplot_showhide_colorbar_axes():
    fields = ("density", "temperature")
    units = (
        "g/cm**3",
        "K",
    )
    ds = fake_random_ds(16, fields=fields, units=units)
    ad = ds.all_data()
    plot = yt.PhasePlot(ad, ("gas", "density"), ("gas", "temperature"), ("gas", "mass"))

    # make sure we can hide colorbar
    plot.hide_colorbar()
    with tempfile.NamedTemporaryFile(suffix="png") as f1:
        plot.save(f1.name)

    # make sure we can show colorbar
    plot.show_colorbar()
    with tempfile.NamedTemporaryFile(suffix="png") as f2:
        plot.save(f2.name)

    # make sure we can hide axes
    plot.hide_axes()
    with tempfile.NamedTemporaryFile(suffix="png") as f3:
        plot.save(f3.name)

    # make sure we can show axes
    plot.show_axes()
    with tempfile.NamedTemporaryFile(suffix="png") as f4:
        plot.save(f4.name)



cimport cython
cimport numpy as np

from yt.utilities.lib.bounded_priority_queue cimport BoundedPriorityQueue
from yt.utilities.lib.cykdtree.kdtree cimport KDTree, uint64_t


cdef struct axes_range:
    int start
    int stop
    int step

cdef int set_axes_range(axes_range *axes, int skipaxis)

cdef int find_neighbors(np.float64_t * pos, np.float64_t[:, ::1] tree_positions,
                        BoundedPriorityQueue queue, KDTree * c_tree,
                        uint64_t skipidx, axes_range * axes) nogil except -1



from nose.plugins.attrib import attr

from yt.config import ytcfg
from yt.testing import (
    ANSWER_TEST_TAG,
    fake_hexahedral_ds,
    fake_tetrahedral_ds,
    requires_module,
)
from yt.utilities.answer_testing.framework import (
    GenericImageTest,
    data_dir_load,
    requires_ds,
)
from yt.visualization.volume_rendering.api import MeshSource, Scene, create_scene


def compare(ds, im, test_prefix, test_name=None, decimals=12):
    def mesh_render_image_func(filename_prefix):
        return im.write_image(filename_prefix)

    mesh_render_image_func.__name__ = f"func_{test_prefix}"
    test = GenericImageTest(ds, mesh_render_image_func, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test


def surface_mesh_render():
    images = []

    ds = fake_tetrahedral_ds()
    for field in ds.field_list:
        if field[0] == "all":
            continue
        sc = Scene()
        sc.add_source(MeshSource(ds, field))
        sc.add_camera()
        im = sc.render()
        images.append(im)

    ds = fake_hexahedral_ds()
    for field in ds.field_list:
        if field[0] == "all":
            continue
        sc = Scene()
        sc.add_source(MeshSource(ds, field))
        sc.add_camera()
        im = sc.render()
        images.append(im)

    return images


@requires_module("pyembree")
def test_surface_mesh_render_pyembree():
    ytcfg["yt", "ray_tracing_engine"] = "embree"
    surface_mesh_render()


def test_surface_mesh_render():
    ytcfg["yt", "ray_tracing_engine"] = "yt"
    surface_mesh_render()


@attr(ANSWER_TEST_TAG)
def test_fake_hexahedral_ds_render():
    ds = fake_hexahedral_ds()
    field_list = [("connect1", "elem"), ("connect1", "test")]
    for field in field_list:
        sc = create_scene(ds, field)
        im = sc.render()
        test_prefix = f"yt_render_fake_hexahedral_{field[0]}_{field[1]}"
        yield compare(
            ds, im, test_prefix=test_prefix, test_name="fake_hexahedral_ds_render"
        )


hex8 = "MOOSE_sample_data/out.e-s010"
hex8_fields = [("connect1", "diffused"), ("connect2", "convected")]


def hex8_render(engine, field):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(hex8, kwargs={"step": -1})
    sc = create_scene(ds, field)
    im = sc.render()
    return compare(ds, im, f"{engine}_render_answers_hex8_{field[0]}_{field[1]}")


@requires_ds(hex8)
@requires_module("pyembree")
def test_hex8_render_pyembree():
    for field in hex8_fields:
        yield hex8_render("embree", field)


@requires_ds(hex8)
def test_hex8_render():
    for field in hex8_fields:
        yield hex8_render("yt", field)


tet4 = "MOOSE_sample_data/high_order_elems_tet4_refine_out.e"
tet4_fields = [("connect1", "u")]


def tet4_render(engine, field):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(tet4, kwargs={"step": -1})
    sc = create_scene(ds, field)
    im = sc.render()
    return compare(ds, im, f"{engine}_render_answers_tet4_{field[0]}_{field[1]}")


@requires_ds(tet4)
@requires_module("pyembree")
def test_tet4_render_pyembree():
    for field in tet4_fields:
        yield tet4_render("embree", field)


@requires_ds(tet4)
def test_tet4_render():
    for field in tet4_fields:
        yield tet4_render("yt", field)


hex20 = "MOOSE_sample_data/mps_out.e"
hex20_fields = [("connect2", "temp")]


def hex20_render(engine, field):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(hex20, kwargs={"step": -1})
    sc = create_scene(ds, field)
    im = sc.render()
    return compare(ds, im, f"{engine}_render_answers_hex20_{field[0]}_{field[1]}")


@requires_ds(hex20)
@requires_module("pyembree")
def test_hex20_render_pyembree():
    for field in hex20_fields:
        yield hex20_render("embree", field)


@requires_ds(hex20)
def test_hex20_render():
    for field in hex20_fields:
        yield hex20_render("yt", field)


wedge6 = "MOOSE_sample_data/wedge_out.e"
wedge6_fields = [("connect1", "diffused")]


def wedge6_render(engine, field):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(wedge6, kwargs={"step": -1})
    sc = create_scene(ds, field)
    im = sc.render()
    return compare(ds, im, f"{engine}_render_answers_wedge6_{field[0]}_{field[1]}")


@requires_ds(wedge6)
@requires_module("pyembree")
def test_wedge6_render_pyembree():
    for field in wedge6_fields:
        yield wedge6_render("embree", field)


@requires_ds(wedge6)
def test_wedge6_render():
    for field in wedge6_fields:
        yield wedge6_render("yt", field)


tet10 = "SecondOrderTets/tet10_unstructured_out.e"
tet10_fields = [("connect1", "uz")]


def tet10_render(engine, field):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(tet10, kwargs={"step": -1})
    sc = create_scene(ds, field)
    ms = sc.get_source(0)
    ms.color_bounds = (-0.01, 0.2)
    im = sc.render()
    return compare(ds, im, f"{engine}_render_answers_tet10_{field[0]}_{field[1]}")


@requires_ds(tet10)
@requires_module("pyembree")
def test_tet10_render_pyembree():
    for field in tet10_fields:
        yield tet10_render("embree", field)


@requires_ds(tet10)
def test_tet10_render():
    for field in tet10_fields:
        yield tet10_render("yt", field)


def perspective_mesh_render(engine):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(hex8)
    sc = create_scene(ds, ("connect2", "diffused"))
    cam = sc.add_camera(ds, lens_type="perspective")
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam_pos = ds.arr([-4.5, 4.5, -4.5], "code_length")
    north_vector = ds.arr([0.0, -1.0, -1.0], "dimensionless")
    cam.set_position(cam_pos, north_vector)
    cam.resolution = (800, 800)
    im = sc.render()
    return compare(ds, im, f"{engine}_perspective_mesh_render")


@requires_ds(hex8)
@requires_module("pyembree")
def test_perspective_mesh_render_pyembree():
    yield perspective_mesh_render("embree")


@requires_ds(hex8)
def test_perspective_mesh_render():
    yield perspective_mesh_render("yt")


def composite_mesh_render(engine):
    ytcfg["yt", "ray_tracing_engine"] = engine
    ds = data_dir_load(hex8)
    sc = Scene()
    cam = sc.add_camera(ds)
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam.set_position(
        ds.arr([-3.0, 3.0, -3.0], "code_length"),
        ds.arr([0.0, -1.0, 0.0], "dimensionless"),
    )
    cam.set_width = ds.arr([8.0, 8.0, 8.0], "code_length")
    cam.resolution = (800, 800)
    ms1 = MeshSource(ds, ("connect1", "diffused"))
    ms2 = MeshSource(ds, ("connect2", "diffused"))
    sc.add_source(ms1)
    sc.add_source(ms2)
    im = sc.render()
    return compare(ds, im, f"{engine}_composite_mesh_render")


@requires_ds(hex8)
@requires_module("pyembree")
def test_composite_mesh_render_pyembree():
    yield composite_mesh_render("embree")


@requires_ds(hex8)
def test_composite_mesh_render():
    yield composite_mesh_render("yt")



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import yt
from yt.visualization.api import Streamlines

# Load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Define c: the center of the box, N: the number of streamlines,
# scale: the spatial scale of the streamlines relative to the boxsize,
# and then pos: the random positions of the streamlines.
c = ds.arr([0.5] * 3, "code_length")
N = 30
scale = ds.quan(15, "kpc").in_units("code_length")  # 15 kpc in code units
pos_dx = np.random.random((N, 3)) * scale - scale / 2.0
pos = c + pos_dx

# Create the streamlines from these positions with the velocity fields as the
# fields to be traced
streamlines = Streamlines(
    ds,
    pos,
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
    ("gas", "velocity_z"),
    length=1.0,
)
streamlines.integrate_through_volume()

# Create a 3D matplotlib figure for visualizing the streamlines
fig = plt.figure()
ax = Axes3D(fig, auto_add_to_figure=False)
fig.add_axes(ax)

# Trace the streamlines through the volume of the 3D figure
for stream in streamlines.streamlines:
    stream = stream[np.all(stream != 0.0, axis=1)]

    # Make the colors of each stream vary continuously from blue to red
    # from low-x to high-x of the stream start position (each color is R, G, B)
    # can omit and just set streamline colors to a fixed color
    x_start_pos = ds.arr(stream[0, 0], "code_length")
    x_start_pos -= ds.arr(0.5, "code_length")
    x_start_pos /= scale
    x_start_pos += 0.5
    color = np.array([x_start_pos, 0, 1 - x_start_pos])

    # Plot the stream in 3D
    ax.plot3D(stream[:, 0], stream[:, 1], stream[:, 2], alpha=0.3, color=color)

# Create a sphere object centered on the highest density point in the simulation
# with radius = 1 Mpc
sphere = ds.sphere("max", (1.0, "Mpc"))

# Identify the isodensity surface in this sphere with density = 1e-24 g/cm^3
surface = ds.surface(sphere, ("gas", "density"), 1e-24)

# Color this isodensity surface according to the log of the temperature field
colors = yt.apply_colormap(np.log10(surface[("gas", "temperature")]), cmap_name="hot")

# Render this surface
p3dc = Poly3DCollection(surface.triangles, linewidth=0.0)
colors = colors[0, :, :] / 255.0  # scale to [0,1]
colors[:, 3] = 0.3  # alpha = 0.3
p3dc.set_facecolors(colors)
ax.add_collection(p3dc)

# Save the figure
plt.savefig("streamlines_isocontour.png")



cdef class GridSelector(SelectorObject):
    cdef object ind

    def __init__(self, dobj):
        self.ind = dobj.id - dobj._id_offset

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        cdef int ng = left_edges.shape[0]
        cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
        gridi[self.ind] = 1
        return gridi.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask_regular_grid(self, gobj):
        return np.ones(gobj.ActiveDimensions, dtype='bool')

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        return 1

    cdef int select_point(self, np.float64_t pos[3]) nogil:
        # we apparently don't check if the point actually lies in the grid..
        return 1

    def _hash_vals(self):
        return (self.ind,)

grid_selector = GridSelector



"""Tests for non-cartesian ray tracers."""
import numpy as np

from yt.testing import amrspace, assert_equal, assert_less_equal, assert_true
from yt.utilities.lib.alt_ray_tracers import _cyl2cart, cylindrical_ray_trace

left_grid = right_grid = amr_levels = center_grid = data = None


def setup():
    # set up some sample cylindrical grid data, radiating out from center
    global left_grid, right_grid, amr_levels, center_grid, data
    np.seterr(all="ignore")
    l1, r1, lvl1 = amrspace([0.0, 1.0, 0.0, -1.0, 0.0, 2 * np.pi], levels=(7, 7, 0))
    l2, r2, lvl2 = amrspace([0.0, 1.0, 0.0, 1.0, 0.0, 2 * np.pi], levels=(7, 7, 0))
    left_grid = np.concatenate([l1, l2], axis=0)
    right_grid = np.concatenate([r1, r2], axis=0)
    amr_levels = np.concatenate([lvl1, lvl2], axis=0)
    center_grid = (left_grid + right_grid) / 2.0
    data = np.cos(np.sqrt(np.sum(center_grid[:, :2] ** 2, axis=1))) ** 2  # cos^2


point_pairs = np.array(
    [
        # p1               p2
        ([0.5, -1.0, 0.0], [1.0, 1.0, 0.75 * np.pi]),  # Everything different
        ([0.5, -1.0, 0.0], [0.5, 1.0, 0.75 * np.pi]),  # r same
        ([0.5, -1.0, 0.0], [0.5, 1.0, np.pi]),  # diagonal through z-axis
        # straight through z-axis
        ([0.5, 0.0, 0.0], [0.5, 0.0, np.pi]),
        # ([0.5, 0.0, np.pi*3/2 + 0.0], [0.5, 0.0, np.pi*3/2 + np.pi]),
        # ([0.5, 0.0, np.pi/2 + 0.0], [0.5, 0.0, np.pi/2 + np.pi]),
        # ([0.5, 0.0, np.pi + 0.0], [0.5, 0.0, np.pi + np.pi]),
        # const z, not through z-axis
        ([0.5, 0.1, 0.0], [0.5, 0.1, 0.75 * np.pi]),
        # ([0.5, 0.1, np.pi + 0.0], [0.5, 0.1, np.pi + 0.75*np.pi]),
        # ([0.5, 0.1, np.pi*3/2 + 0.0], [0.5, 0.1, np.pi*3/2 + 0.75*np.pi]),
        # ([0.5, 0.1, np.pi/2 + 0.0], [0.5, 0.1, np.pi/2 + 0.75*np.pi]),
        # ([0.5, 0.1, 2*np.pi + 0.0], [0.5, 0.1, 2*np.pi + 0.75*np.pi]),
        # ([0.5, 0.1, np.pi/4 + 0.0], [0.5, 0.1, np.pi/4 + 0.75*np.pi]),
        # ([0.5, 0.1, np.pi*3/8 + 0.0], [0.5, 0.1, np.pi*3/8 + 0.75*np.pi]),
        (
            [0.5, -1.0, 0.75 * np.pi],
            [1.0, 1.0, 0.75 * np.pi],
        ),  # r,z different - theta same
        ([0.5, -1.0, 0.75 * np.pi], [0.5, 1.0, 0.75 * np.pi]),  # z-axis parallel
        ([0.0, -1.0, 0.0], [0.0, 1.0, 0.0]),  # z-axis itself
    ]
)


def check_monotonic_inc(arr):
    assert_true(np.all(0.0 <= (arr[1:] - arr[:-1])))


def check_bounds(arr, blower, bupper):
    assert_true(np.all(blower <= arr))
    assert_true(np.all(bupper >= arr))


def test_cylindrical_ray_trace():
    for pair in point_pairs:
        p1, p2 = pair
        p1cart, p2cart = _cyl2cart(pair)
        pathlen = np.sqrt(np.sum((p2cart - p1cart) ** 2))

        t, s, rztheta, inds = cylindrical_ray_trace(p1, p2, left_grid, right_grid)
        npoints = len(t)

        check_monotonic_inc(t)
        assert_less_equal(0.0, t[0])
        assert_less_equal(t[-1], 1.0)

        check_monotonic_inc(s)
        assert_less_equal(0.0, s[0])
        assert_less_equal(s[-1], pathlen)
        assert_equal(npoints, len(s))

        assert_equal((npoints, 3), rztheta.shape)
        check_bounds(rztheta[:, 0], 0.0, 1.0)
        check_bounds(rztheta[:, 1], -1.0, 1.0)
        check_bounds(rztheta[:, 2], 0.0, 2 * np.pi)
        check_monotonic_inc(rztheta[:, 2])

        assert_equal(npoints, len(inds))
        check_bounds(inds, 0, len(left_grid) - 1)



import numpy as np

from yt.frontends.athena_pp.api import AthenaPPDataset
from yt.loaders import load
from yt.testing import (
    assert_allclose_units,
    assert_equal,
    disable_dataset_cache,
    requires_file,
    units_override_check,
)
from yt.units import dimensions
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

# Deactivating this problematic test until the dataset type can be
# handled properly, see https://github.com/yt-project/yt/issues/3619
"""
_fields_disk = ("density", "velocity_r")

disk = "KeplerianDisk/disk.out1.00000.athdf"


@requires_ds(disk)
def test_disk():
    ds = data_dir_load(disk)
    assert_equal(str(ds), "disk.out1.00000")
    dd = ds.all_data()
    vol = (ds.domain_right_edge[0] ** 3 - ds.domain_left_edge[0] ** 3) / 3.0
    vol *= np.cos(ds.domain_left_edge[1]) - np.cos(ds.domain_right_edge[1])
    vol *= ds.domain_right_edge[2].v - ds.domain_left_edge[2].v
    assert_allclose(dd.quantities.total_quantity(("gas", "cell_volume")), vol)
    for field in _fields_disk:

        def field_func(name):
            return dd[field]

        yield GenericArrayTest(ds, field_func, args=[field])
"""

_fields_AM06 = ("temperature", "density", "velocity_magnitude", "magnetic_field_x")

AM06 = "AM06/AM06.out1.00400.athdf"


@requires_ds(AM06)
def test_AM06():
    ds = data_dir_load(AM06)
    assert_equal(str(ds), "AM06.out1.00400")
    for test in small_patch_amr(ds, _fields_AM06):
        test_AM06.__name__ = test.description
        yield test


uo_AM06 = {
    "length_unit": (1.0, "kpc"),
    "mass_unit": (1.0, "Msun"),
    "time_unit": (1.0, "Myr"),
}


@requires_file(AM06)
def test_AM06_override():
    # verify that overriding units causes derived unit values to be updated.
    # see issue #1259
    ds = load(AM06, units_override=uo_AM06)
    assert_equal(float(ds.magnetic_unit.in_units("gauss")), 9.01735778342523e-08)


@requires_file(AM06)
def test_units_override():
    units_override_check(AM06)


@requires_file(AM06)
def test_AthenaPPDataset():
    assert isinstance(data_dir_load(AM06), AthenaPPDataset)


@requires_file(AM06)
def test_magnetic_units():
    ds = load(AM06, unit_system="code")
    assert ds.magnetic_unit.units.dimensions == dimensions.magnetic_field_cgs
    assert (ds.magnetic_unit**2).units.dimensions == dimensions.pressure


@requires_file(AM06)
@disable_dataset_cache
def test_mag_factor():
    ds1 = load(AM06, units_override=uo_AM06, magnetic_normalization="gaussian")
    assert ds1.magnetic_unit == np.sqrt(
        4.0 * np.pi * ds1.mass_unit / (ds1.time_unit**2 * ds1.length_unit)
    )
    sp1 = ds1.sphere("c", (100.0, "kpc"))
    pB1a = (
        sp1["athena_pp", "Bcc1"] ** 2
        + sp1["athena_pp", "Bcc2"] ** 2
        + sp1["athena_pp", "Bcc3"] ** 2
    ) / (8.0 * np.pi)
    pB1b = (
        sp1["gas", "magnetic_field_x"] ** 2
        + sp1["gas", "magnetic_field_y"] ** 2
        + sp1["gas", "magnetic_field_z"] ** 2
    ) / (8.0 * np.pi)
    pB1a.convert_to_units("dyn/cm**2")
    pB1b.convert_to_units("dyn/cm**2")
    assert_allclose_units(pB1a, pB1b)
    assert_allclose_units(pB1a, sp1["magnetic_pressure"])
    ds2 = load(AM06, units_override=uo_AM06, magnetic_normalization="lorentz_heaviside")
    assert ds2.magnetic_unit == np.sqrt(
        ds2.mass_unit / (ds2.time_unit**2 * ds2.length_unit)
    )
    sp2 = ds2.sphere("c", (100.0, "kpc"))
    pB2a = (
        sp2["athena_pp", "Bcc1"] ** 2
        + sp2["athena_pp", "Bcc2"] ** 2
        + sp2["athena_pp", "Bcc3"] ** 2
    ) / 2.0
    pB2b = (
        sp2["gas", "magnetic_field_x"] ** 2
        + sp2["gas", "magnetic_field_y"] ** 2
        + sp2["gas", "magnetic_field_z"] ** 2
    ) / 2.0
    pB2a.convert_to_units("dyn/cm**2")
    pB2b.convert_to_units("dyn/cm**2")
    assert_allclose_units(pB2a, pB2b)
    assert_allclose_units(pB2a, sp2["magnetic_pressure"])
    assert_allclose_units(pB1a, pB2a)



import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid

import yt

fns = [
    "Enzo_64/DD0005/data0005",
    "Enzo_64/DD0015/data0015",
    "Enzo_64/DD0025/data0025",
    "Enzo_64/DD0035/data0035",
]

fig = plt.figure()

# See http://matplotlib.org/mpl_toolkits/axes_grid/api/axes_grid_api.html
# These choices of keyword arguments produce a four panel plot with a single
# shared narrow colorbar on the right hand side of the multipanel plot. Axes
# labels are drawn for all plots since we're slicing along different directions
# for each plot.
grid = AxesGrid(
    fig,
    (0.075, 0.075, 0.85, 0.85),
    nrows_ncols=(2, 2),
    axes_pad=0.05,
    label_mode="L",
    share_all=True,
    cbar_location="right",
    cbar_mode="single",
    cbar_size="3%",
    cbar_pad="0%",
)

for i, fn in enumerate(fns):
    # Load the data and create a single plot
    ds = yt.load(fn)  # load data
    p = yt.ProjectionPlot(ds, "z", ("gas", "density"), width=(55, "Mpccm"))

    # Ensure the colorbar limits match for all plots
    p.set_zlim(("gas", "density"), 1e-4, 1e-2)

    # This forces the ProjectionPlot to redraw itself on the AxesGrid axes.
    plot = p.plots[("gas", "density")]
    plot.figure = fig
    plot.axes = grid[i].axes
    plot.cax = grid.cbar_axes[i]

    # Finally, this actually redraws the plot.
    p._setup_plots()

plt.savefig("multiplot_2x2_time_series.png")



import os
import sys

import numpy as np

from yt.data_objects.static_output import ParticleDataset, ParticleFile
from yt.funcs import setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _requests as requests
from yt.utilities.sdf import HTTPSDFRead, SDFIndex, SDFRead

from .fields import SDFFieldInfo

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from yt._maintenance.backports import cached_property
# currently specified by units_2HOT == 2 in header
# in future will read directly from file
units_2HOT_v2_length = 3.08567802e21
units_2HOT_v2_mass = 1.98892e43
units_2HOT_v2_time = 3.1558149984e16


class SDFFile(ParticleFile):
    pass


class SDFDataset(ParticleDataset):
    _index_class = ParticleIndex
    _file_class = SDFFile
    _field_info_class = SDFFieldInfo
    _particle_mass_name = None
    _particle_coordinates_name = None
    _particle_velocity_name = None
    _skip_cache = True
    _subspace = False

    def __init__(
        self,
        filename,
        dataset_type="sdf_particles",
        index_order=None,
        index_filename=None,
        bounding_box=None,
        sdf_header=None,
        midx_filename=None,
        midx_header=None,
        midx_level=None,
        field_map=None,
        units_override=None,
        unit_system="cgs",
    ):
        if bounding_box is not None:
            # This ensures that we know a bounding box has been applied
            self._domain_override = True
            self._subspace = True
            bbox = np.array(bounding_box, dtype="float64")
            if bbox.shape == (2, 3):
                bbox = bbox.transpose()
            self.domain_left_edge = bbox[:, 0]
            self.domain_right_edge = bbox[:, 1]
        else:
            self.domain_left_edge = self.domain_right_edge = None
        self.sdf_header = sdf_header
        self.midx_filename = midx_filename
        self.midx_header = midx_header
        self.midx_level = midx_level
        if field_map is None:
            field_map = {}
        self._field_map = field_map
        prefix = ""
        if self.midx_filename is not None:
            prefix += "midx_"
        if filename.startswith("http"):
            prefix += "http_"
        dataset_type = prefix + "sdf_particles"
        super().__init__(
            filename,
            dataset_type=dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            index_order=index_order,
            index_filename=index_filename,
        )

    def _parse_parameter_file(self):
        if self.parameter_filename.startswith("http"):
            sdf_class = HTTPSDFRead
        else:
            sdf_class = SDFRead
        self.sdf_container = sdf_class(self.parameter_filename, header=self.sdf_header)

        # Reference
        self.parameters = self.sdf_container.parameters
        self.dimensionality = 3
        self.refine_by = 2

        if self.domain_left_edge is None or self.domain_right_edge is None:
            R0 = self.parameters["R0"]
            if "offset_center" in self.parameters and self.parameters["offset_center"]:
                self.domain_left_edge = np.array([0, 0, 0], dtype=np.float64)
                self.domain_right_edge = np.array(
                    [2.0 * self.parameters.get(f"R{ax}", R0) for ax in "xyz"],
                    dtype=np.float64,
                )
            else:
                self.domain_left_edge = np.array(
                    [-self.parameters.get(f"R{ax}", R0) for ax in "xyz"],
                    dtype=np.float64,
                )
                self.domain_right_edge = np.array(
                    [+self.parameters.get(f"R{ax}", R0) for ax in "xyz"],
                    dtype=np.float64,
                )
            self.domain_left_edge *= self.parameters.get("a", 1.0)
            self.domain_right_edge *= self.parameters.get("a", 1.0)

        self.domain_dimensions = np.ones(3, "int32")
        if self.parameters.get("do_periodic", False):
            self._periodicity = (True, True, True)
        else:
            self._periodicity = (False, False, False)

        self.cosmological_simulation = 1

        self.current_redshift = self.parameters.get("redshift", 0.0)
        self.omega_lambda = self.parameters["Omega0_lambda"]
        self.omega_matter = self.parameters["Omega0_m"]
        if "Omega0_fld" in self.parameters:
            self.omega_lambda += self.parameters["Omega0_fld"]
        if "Omega0_r" in self.parameters:
            # not correct, but most codes can't handle Omega0_r
            self.omega_matter += self.parameters["Omega0_r"]
        self.hubble_constant = self.parameters["h_100"]
        self.current_time = units_2HOT_v2_time * self.parameters.get("tpos", 0.0)
        mylog.info("Calculating time to be %0.3e seconds", self.current_time)
        self.filename_template = self.parameter_filename
        self.file_count = 1

    @cached_property
    def midx(self):
        if self.midx_filename is None:
            raise RuntimeError("SDF index0 file not supplied in load.")

        if "http" in self.midx_filename:
            sdf_class = HTTPSDFRead
        else:
            sdf_class = SDFRead
        indexdata = sdf_class(self.midx_filename, header=self.midx_header)
        return SDFIndex(self.sdf_container, indexdata, level=self.midx_level)

    def _set_code_unit_attributes(self):
        setdefaultattr(
            self,
            "length_unit",
            self.quan(1.0, self.parameters.get("length_unit", "kpc")),
        )
        setdefaultattr(
            self,
            "velocity_unit",
            self.quan(1.0, self.parameters.get("velocity_unit", "kpc/Gyr")),
        )
        setdefaultattr(
            self, "time_unit", self.quan(1.0, self.parameters.get("time_unit", "Gyr"))
        )
        mass_unit = self.parameters.get("mass_unit", "1e10 Msun")
        if " " in mass_unit:
            factor, unit = mass_unit.split(" ")
        else:
            factor = 1.0
            unit = mass_unit
        setdefaultattr(self, "mass_unit", self.quan(float(factor), unit))

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        sdf_header = kwargs.get("sdf_header", filename)
        if sdf_header.startswith("http"):
            try:
                hreq = requests.get(sdf_header, stream=True)
            except ImportError:
                # requests is not installed
                return False
            if hreq.status_code != 200:
                return False
            # Grab a whole 4k page.
            line = next(hreq.iter_content(4096))
        elif os.path.isfile(sdf_header):
            with open(sdf_header, encoding="ISO-8859-1") as f:
                line = f.read(10).strip()
        else:
            return False
        return line.startswith("# SDF")



import inspect
from textwrap import TextWrapper

import yt

ds = yt.load("RD0005-mine/RedshiftOutput0005")

output = open("source/visualizing/_cb_docstrings.inc", "w")

template = """

.. function:: %(clsname)s%(sig)s:

   (This is a proxy for :class:`~%(clsproxy)s`.)

%(docstring)s

"""

tw = TextWrapper(initial_indent="   ", subsequent_indent="   ", width=60)


def write_docstring(f, name, cls):
    if not hasattr(cls, "_type_name") or cls._type_name is None:
        return
    for clsi in inspect.getmro(cls):
        docstring = inspect.getdoc(clsi.__init__)
        if docstring is not None:
            break
    clsname = cls._type_name
    sig = inspect.formatargspec(*inspect.getargspec(cls.__init__))
    sig = sig.replace("**kwargs", "**field_parameters")
    clsproxy = f"yt.visualization.plot_modifications.{cls.__name__}"
    # docstring = "\n".join(["   %s" % line for line in docstring.split("\n")])
    # print(docstring)
    f.write(
        template
        % dict(
            clsname=clsname,
            sig=sig,
            clsproxy=clsproxy,
            docstring="\n".join(tw.wrap(docstring)),
        )
    )
    # docstring = docstring))


for n, c in sorted(yt.visualization.api.callback_registry.items()):
    write_docstring(output, n, c)
    print(f".. autoclass:: yt.visualization.plot_modifications.{n}")
    print("   :members:")
    print()



import abc
from typing import Tuple

domain_context_registry = {}


class DomainContext(abc.ABC):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            domain_context_registry[name] = cls

    _known_fluid_fields: Tuple[Tuple[str, str], ...]
    _known_particle_fields: Tuple[Tuple[str, str], ...]

    def __init__(self, ds):
        self.ds = ds



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import (
    fix_length,
    validate_3d_array,
    validate_center,
    validate_float,
    validate_object,
    validate_sequence,
)
from yt.units import YTArray
from yt.utilities.exceptions import YTEllipsoidOrdering, YTException, YTSphereTooSmall
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.math_utils import get_rotation_matrix
from yt.utilities.on_demand_imports import _miniball


class YTSphere(YTSelectionContainer3D):
    """
    A sphere of points defined by a *center* and a *radius*.

    Parameters
    ----------
    center : array_like
        The center of the sphere.
    radius : float, width specifier, or YTQuantity
        The radius of the sphere. If passed a float,
        that will be interpreted in code units. Also
        accepts a (radius, unit) tuple or YTQuantity
        instance with units attached.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5, 0.5, 0.5]
    >>> sphere = ds.sphere(c, (1.0, "kpc"))
    """

    _type_name = "sphere"
    _con_args = ("center", "radius")

    def __init__(
        self, center, radius, ds=None, field_parameters=None, data_source=None
    ):
        validate_center(center)
        validate_float(radius)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super().__init__(center, ds, field_parameters, data_source)
        # Unpack the radius, if necessary
        radius = fix_length(radius, self.ds)
        if radius < self.index.get_smallest_dx():
            raise YTSphereTooSmall(
                ds,
                radius.in_units("code_length"),
                self.index.get_smallest_dx().in_units("code_length"),
            )
        self.set_field_parameter("radius", radius)
        self.set_field_parameter("center", self.center)
        self.radius = radius

    def _get_bbox(self):
        """
        Return the minimum bounding box for the sphere.
        """
        return -self.radius + self.center, self.radius + self.center


class YTMinimalSphere(YTSelectionContainer3D):
    """
    Build the smallest sphere that encompasses a set of points.

    Parameters
    ----------
    points : YTArray
        The points that the sphere will contain.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("output_00080/info_00080.txt")
    >>> points = ds.r["particle_position"]
    >>> sphere = ds.minimal_sphere(points)
    """

    _type_name = "sphere"
    _override_selector_name = "minimal_sphere"
    _con_args = ("center", "radius")

    def __init__(self, points, ds=None, field_parameters=None, data_source=None):
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        validate_object(points, YTArray)

        points = fix_length(points, ds)
        if len(points) < 2:
            raise YTException(
                f"Not enough points. Expected at least 2, got {len(points)}"
            )
        mylog.debug("Building minimal sphere around points.")
        mb = _miniball.Miniball(points)
        if not mb.is_valid():
            raise YTException("Could not build valid sphere around points.")

        center = ds.arr(mb.center(), points.units)
        radius = ds.quan(np.sqrt(mb.squared_radius()), points.units)
        super().__init__(center, ds, field_parameters, data_source)
        self.set_field_parameter("radius", radius)
        self.set_field_parameter("center", self.center)
        self.radius = radius


class YTEllipsoid(YTSelectionContainer3D):
    """
    By providing a *center*,*A*,*B*,*C*,*e0*,*tilt* we
    can define a ellipsoid of any proportion.  Only cells whose
    centers are within the ellipsoid will be selected.

    Parameters
    ----------
    center : array_like
        The center of the ellipsoid.
    A : float
        The magnitude of the largest axis (semi-major) of the ellipsoid.
    B : float
        The magnitude of the medium axis (semi-medium) of the ellipsoid.
    C : float
        The magnitude of the smallest axis (semi-minor) of the ellipsoid.
    e0 : array_like (automatically normalized)
        the direction of the largest semi-major axis of the ellipsoid
    tilt : float
        After the rotation about the z-axis to align e0 to x in the x-y
        plane, and then rotating about the y-axis to align e0 completely
        to the x-axis, tilt is the angle in radians remaining to
        rotate about the x-axis to align both e1 to the y-axis and e2 to
        the z-axis.
    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5, 0.5, 0.5]
    >>> ell = ds.ellipsoid(c, 0.1, 0.1, 0.1, np.array([0.1, 0.1, 0.1]), 0.2)
    """

    _type_name = "ellipsoid"
    _con_args = ("center", "_A", "_B", "_C", "_e0", "_tilt")

    def __init__(
        self,
        center,
        A,
        B,
        C,
        e0,
        tilt,
        fields=None,
        ds=None,
        field_parameters=None,
        data_source=None,
    ):
        validate_center(center)
        validate_float(A)
        validate_float(B)
        validate_float(C)
        validate_3d_array(e0)
        validate_float(tilt)
        validate_sequence(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters, data_source)
        # make sure the magnitudes of semi-major axes are in order
        if A < B or B < C:
            raise YTEllipsoidOrdering(ds, A, B, C)
        # make sure the smallest side is not smaller than dx
        self._A = self.ds.quan(A, "code_length")
        self._B = self.ds.quan(B, "code_length")
        self._C = self.ds.quan(C, "code_length")
        if self._C < self.index.get_smallest_dx():
            raise YTSphereTooSmall(self.ds, self._C, self.index.get_smallest_dx())
        self._e0 = e0 = e0 / (e0**2.0).sum() ** 0.5
        self._tilt = tilt

        # find the t1 angle needed to rotate about z axis to align e0 to x
        t1 = np.arctan(e0[1] / e0[0])
        # rotate e0 by -t1
        RZ = get_rotation_matrix(t1, (0, 0, 1)).transpose()
        r1 = (e0 * RZ).sum(axis=1)
        # find the t2 angle needed to rotate about y axis to align e0 to x
        t2 = np.arctan(-r1[2] / r1[0])
        """
        calculate the original e1
        given the tilt about the x axis when e0 was aligned
        to x after t1, t2 rotations about z, y
        """
        RX = get_rotation_matrix(-tilt, (1, 0, 0)).transpose()
        RY = get_rotation_matrix(-t2, (0, 1, 0)).transpose()
        RZ = get_rotation_matrix(-t1, (0, 0, 1)).transpose()
        e1 = ((0, 1, 0) * RX).sum(axis=1)
        e1 = (e1 * RY).sum(axis=1)
        e1 = (e1 * RZ).sum(axis=1)
        e2 = np.cross(e0, e1)

        self._e1 = e1
        self._e2 = e2

        self.set_field_parameter("A", A)
        self.set_field_parameter("B", B)
        self.set_field_parameter("C", C)
        self.set_field_parameter("e0", e0)
        self.set_field_parameter("e1", e1)
        self.set_field_parameter("e2", e2)

    def _get_bbox(self):
        """
        Get the bounding box for the ellipsoid. NOTE that in this case
        it is not the *minimum* bounding box.
        """
        radius = self.ds.arr(np.max([self._A, self._B, self._C]), "code_length")
        return -radius + self.center, radius + self.center



import yt

# Load the dataset.
ds = yt.load("GalaxyClusterMerger/fiducial_1to3_b0.273d_hdf5_plt_cnt_0175")

# Create projections of the gas density (non-weighted line integrals).

yt.ProjectionPlot(ds, "x", ("gas", "density")).save()
yt.ProjectionPlot(ds, "y", ("gas", "density")).save()
yt.ProjectionPlot(ds, "z", ("gas", "density")).save()



import re
import subprocess

from docutils import statemachine
from docutils.parsers.rst import Directive


def setup(app):
    app.add_directive("config_help", GetConfigHelp)
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    retdict = dict(version="1.0", parallel_read_safe=True, parallel_write_safe=True)

    return retdict


class GetConfigHelp(Directive):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    def run(self):
        rst_file = self.state_machine.document.attributes["source"]
        data = (
            subprocess.check_output(self.arguments[0].split(" ") + ["-h"])
            .decode("utf8")
            .split("\n")
        )
        ind = next(
            (i for i, val in enumerate(data) if re.match(r"\s{0,3}\{.*\}\s*$", val))
        )
        lines = [".. code-block:: none", ""] + data[ind + 1 :]
        self.state_machine.insert_input(
            statemachine.string2lines("\n".join(lines)), rst_file
        )
        return []



from collections import OrderedDict

from yt.frontends.tipsy.api import TipsyDataset
from yt.testing import ParticleSelectionComparison, requires_file
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    nbody_answer,
    requires_ds,
    sph_answer,
)

_fields = OrderedDict(
    [
        (("all", "particle_mass"), None),
        (("all", "particle_ones"), None),
        (("all", "particle_velocity_x"), ("all", "particle_mass")),
        (("all", "particle_velocity_y"), ("all", "particle_mass")),
        (("all", "particle_velocity_z"), ("all", "particle_mass")),
    ]
)

pkdgrav = "halo1e11_run1.00400/halo1e11_run1.00400"
pkdgrav_cosmology_parameters = dict(
    current_redshift=0.0, omega_lambda=0.728, omega_matter=0.272, hubble_constant=0.702
)
pkdgrav_kwargs = dict(
    field_dtypes={"Coordinates": "d"},
    cosmology_parameters=pkdgrav_cosmology_parameters,
    unit_base={"length": (60.0, "Mpccm/h")},
)


@requires_ds(pkdgrav, big_data=True, file_check=True)
def test_pkdgrav():
    ds = data_dir_load(pkdgrav, TipsyDataset, (), kwargs=pkdgrav_kwargs)
    yield from nbody_answer(ds, "halo1e11_run1.00400", 26847360, _fields)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


gasoline_dmonly = "agora_1e11.00400/agora_1e11.00400"


@requires_ds(gasoline_dmonly, big_data=True, file_check=True)
def test_gasoline_dmonly():
    cosmology_parameters = dict(
        current_redshift=0.0,
        omega_lambda=0.728,
        omega_matter=0.272,
        hubble_constant=0.702,
    )
    kwargs = dict(
        cosmology_parameters=cosmology_parameters,
        unit_base={"length": (60.0, "Mpccm/h")},
    )
    ds = data_dir_load(gasoline_dmonly, TipsyDataset, (), kwargs)
    yield from nbody_answer(ds, "agora_1e11.00400", 10550576, _fields)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


tg_sph_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ("gas", "density")),
        (("gas", "velocity_magnitude"), None),
        (("gas", "Fe_fraction"), None),
    ]
)

tg_nbody_fields = OrderedDict(
    [
        (("Stars", "Metals"), None),
    ]
)

tipsy_gal = "TipsyGalaxy/galaxy.00300"


@requires_ds(tipsy_gal)
def test_tipsy_galaxy():
    ds = data_dir_load(
        tipsy_gal,
        kwargs={"bounding_box": [[-2000, 2000], [-2000, 2000], [-2000, 2000]]},
    )
    # These tests should be re-enabled.  But the holdup is that the region
    # selector does not offset by domain_left_edge, and we have inelegant
    # selection using bboxes.
    # psc = ParticleSelectionComparison(ds)
    # psc.run_defaults()
    for test in sph_answer(ds, "galaxy.00300", 315372, tg_sph_fields):
        test_tipsy_galaxy.__name__ = test.description
        yield test
    for test in nbody_answer(ds, "galaxy.00300", 315372, tg_nbody_fields):
        test_tipsy_galaxy.__name__ = test.description
        yield test


@requires_file(gasoline_dmonly)
@requires_file(pkdgrav)
def test_TipsyDataset():
    assert isinstance(data_dir_load(pkdgrav, kwargs=pkdgrav_kwargs), TipsyDataset)
    assert isinstance(data_dir_load(gasoline_dmonly), TipsyDataset)


@requires_file(tipsy_gal)
def test_tipsy_index():
    ds = data_dir_load(tipsy_gal)
    sl = ds.slice("z", 0.0)
    assert sl["gas", "density"].shape[0] != 0



import yt

# Load the dataset.
ds = yt.load("Enzo_64/DD0030/data0030")

# Make a projection that is the full width of the domain,
# but only 5 Mpc in depth.  This is done by creating a
# region object with this exact geometry and providing it
# as a data_source for the projection.

# Center on the domain center
center = ds.domain_center

# First make the left and right corner of the region based
# on the full domain.
left_corner = ds.domain_left_edge
right_corner = ds.domain_right_edge

# Now adjust the size of the region along the line of sight (x axis).
depth = ds.quan(5.0, "Mpc")
left_corner[0] = center[0] - 0.5 * depth
right_corner[0] = center[0] + 0.5 * depth

# Create the region
region = ds.box(left_corner, right_corner)

# Create a density projection and supply the region we have just created.
# Only cells within the region will be included in the projection.
# Try with another data container, like a sphere or disk.
plot = yt.ProjectionPlot(
    ds, "x", ("gas", "density"), weight_field=("gas", "density"), data_source=region
)

# Save the image with the keyword.
plot.save("Thin_Slice")



import numpy as np
from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, fake_vr_orientation_test_ds
from yt.utilities.answer_testing.framework import (
    GenericImageTest,
    VRImageComparisonTest,
)
from yt.visualization.volume_rendering.api import (
    ColorTransferFunction,
    Scene,
    create_volume_source,
    off_axis_projection,
)


@attr(ANSWER_TEST_TAG)
def test_orientation():
    ds = fake_vr_orientation_test_ds()

    sc = Scene()

    vol = create_volume_source(ds, field=("gas", "density"))
    sc.add_source(vol)

    tf = vol.transfer_function
    tf = ColorTransferFunction((0.1, 1.0))
    tf.sample_colormap(1.0, 0.01, colormap="coolwarm")
    tf.sample_colormap(0.8, 0.01, colormap="coolwarm")
    tf.sample_colormap(0.6, 0.01, colormap="coolwarm")
    tf.sample_colormap(0.3, 0.01, colormap="coolwarm")

    n_frames = 1
    orientations = [[-0.3, -0.1, 0.8]]

    theta = np.pi / n_frames
    test_name = "vr_orientation"

    for lens_type, decimals in [("perspective", 12), ("plane-parallel", 2)]:
        # set a much lower precision for plane-parallel tests, see
        # https://github.com/yt-project/yt/issue/3069
        # https://github.com/yt-project/yt/pull/3068
        # https://github.com/yt-project/yt/pull/3294
        frame = 0

        cam = sc.add_camera(ds, lens_type=lens_type)
        cam.resolution = (1000, 1000)
        cam.position = ds.arr(np.array([-4.0, 0.0, 0.0]), "code_length")
        cam.switch_orientation(
            normal_vector=[1.0, 0.0, 0.0], north_vector=[0.0, 0.0, 1.0]
        )
        cam.set_width(ds.domain_width * 2.0)
        desc = "%s_%04d" % (lens_type, frame)
        test1 = VRImageComparisonTest(sc, ds, desc, decimals)
        test1.answer_name = test_name
        yield test1

        for _ in range(n_frames):
            frame += 1
            center = ds.arr([0, 0, 0], "code_length")
            cam.yaw(theta, rot_center=center)
            desc = "yaw_%s_%04d" % (lens_type, frame)
            test2 = VRImageComparisonTest(sc, ds, desc, decimals)
            test2.answer_name = test_name
            yield test2

        for _ in range(n_frames):
            frame += 1
            theta = np.pi / n_frames
            center = ds.arr([0, 0, 0], "code_length")
            cam.pitch(theta, rot_center=center)
            desc = "pitch_%s_%04d" % (lens_type, frame)
            test3 = VRImageComparisonTest(sc, ds, desc, decimals)
            test3.answer_name = test_name
            yield test3

        for _ in range(n_frames):
            frame += 1
            theta = np.pi / n_frames
            center = ds.arr([0, 0, 0], "code_length")
            cam.roll(theta, rot_center=center)
            desc = "roll_%s_%04d" % (lens_type, frame)
            test4 = VRImageComparisonTest(sc, ds, desc, decimals)
            test4.answer_name = test_name
            yield test4

    center = [0.5, 0.5, 0.5]
    width = [1.0, 1.0, 1.0]

    for i, orientation in enumerate(orientations):
        image = off_axis_projection(
            ds, center, orientation, width, 512, ("gas", "density"), no_ghost=False
        )

        def offaxis_image_func(filename_prefix):
            return image.write_image(filename_prefix)  # noqa: B023

        test5 = GenericImageTest(ds, offaxis_image_func, decimals)
        test5.prefix = f"oap_orientation_{i}"
        test5.answer_name = test_name
        yield test5



from functools import partial


# This is the metric for Minkowski spacetime in SR
def metric(mu: int, nu: int):
    # This assumes the -+++ signature
    if (mu, nu) == (0, 0):
        return -1
    elif mu == nu:
        return 1
    else:
        return 0


def setup_stress_energy_ideal(registry, ftype="gas"):

    ax = ("t",) + registry.ds.coordinates.axis_order
    pc = registry.ds.units.physical_constants
    inv_c2 = 1.0 / (pc.clight * pc.clight)

    def _T(field, data, mu: int, nu: int):
        Umu = data[ftype, f"four_velocity_{ax[mu]}"]
        Unu = data[ftype, f"four_velocity_{ax[nu]}"]
        p = data[ftype, "pressure"]
        e = data[ftype, "thermal_energy_density"]
        rho = data[ftype, "density"]
        return (rho + (e + p) * inv_c2) * Umu * Unu + metric(mu, nu) * p

    for mu in range(5):
        for nu in range(5):
            registry.add_field(
                (ftype, f"T{mu}{nu}"),
                sampling_type="local",
                function=partial(_T, mu=mu, nu=nu),
                units=registry.ds.unit_system["pressure"],
            )




# distutils: libraries = STD_LIBS
"""
Utilities for unstructured and semi-structured meshes



"""


import numpy as np

cimport cython
cimport numpy as np
from libc.stdlib cimport abs, free, malloc

from yt.utilities.lib.fp_utils cimport fclip, fmax, fmin, i64clip, iclip, imax, imin

from yt.units.yt_array import YTArray


cdef extern from "platform_dep.h":
    double rint(double x)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def fill_fcoords(np.ndarray[np.float64_t, ndim=2] coords,
                 np.ndarray[np.int64_t, ndim=2] indices,
                 int offset = 0):
    cdef np.ndarray[np.float64_t, ndim=2] fcoords
    cdef int nc = indices.shape[0]
    cdef int nv = indices.shape[1]
    cdef np.float64_t pos[3]
    cdef int i, j, k
    fcoords = np.empty((nc, 3), dtype="float64")
    for i in range(nc):
        for j in range(3):
            pos[j] = 0.0
        for j in range(nv):
            for k in range(3):
                pos[k] += coords[indices[i, j] - offset, k]
        for j in range(3):
            pos[j] /= nv
            fcoords[i, j] = pos[j]
    return fcoords

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def fill_fwidths(np.ndarray[np.float64_t, ndim=2] coords,
                 np.ndarray[np.int64_t, ndim=2] indices,
                 int offset = 0):
    cdef np.ndarray[np.float64_t, ndim=2] fwidths
    cdef int nc = indices.shape[0]
    cdef int nv = indices.shape[1]
    if nv != 8:
        raise NotImplementedError
    cdef np.float64_t LE[3]
    cdef np.float64_t RE[3]
    cdef int i, j, k
    cdef np.float64_t pos
    fwidths = np.empty((nc, 3), dtype="float64")
    for i in range(nc):
        for j in range(3):
            LE[j] = 1e60
            RE[j] = -1e60
        for j in range(nv):
            for k in range(3):
                pos = coords[indices[i, j] - offset, k]
                LE[k] = fmin(pos, LE[k])
                RE[k] = fmax(pos, RE[k])
        for j in range(3):
            fwidths[i, j] = RE[j] - LE[j]
    return fwidths

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def smallest_fwidth(np.ndarray[np.float64_t, ndim=2] coords,
                    np.ndarray[np.int64_t, ndim=2] indices,
                    int offset = 0):
    cdef np.float64_t fwidth = 1e60, pos
    cdef int nc = indices.shape[0]
    cdef int nv = indices.shape[1]
    if nv != 8:
        raise NotImplementedError
    cdef np.float64_t LE[3]
    cdef np.float64_t RE[3]
    cdef int i, j, k
    for i in range(nc):
        for j in range(3):
            LE[j] = 1e60
            RE[j] = -1e60
        for j in range(nv):
            for k in range(3):
                pos = coords[indices[i, j] - offset, k]
                LE[k] = fmin(pos, LE[k])
                RE[k] = fmax(pos, RE[k])
        for j in range(3):
            fwidth = fmin(fwidth, RE[j] - LE[j])
    return fwidth

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def clamp_edges(np.float64_t[:] edge, np.float64_t[:] pleft,
                np.float64_t[:] pdx):
    """Clamp edge to pleft + pdx*n where n is the closest integer

    Note that edge is modified in-place.
    """
    cdef np.float64_t start_index
    cdef np.float64_t integer_index
    cdef np.intp_t shape = edge.shape[0]
    for i in range(shape):
        start_index = (edge[i] - pleft[i]) / pdx[i]
        integer_index = rint(start_index)
        edge[i] = integer_index * pdx[i] + pleft[i]



"""
Particle Deposition onto Octs




"""


cimport numpy as np

import numpy as np

cimport cython
from libc.math cimport sqrt
from libc.stdlib cimport free, malloc, qsort
from oct_container cimport Oct, OctreeContainer

from yt.utilities.lib.distance_queue cimport (
    DistanceQueue,
    Neighbor_compare,
    NeighborList,
    r2dist,
)
from yt.utilities.lib.fp_utils cimport *

from .particle_deposit cimport get_kernel_func, gind, kernel_func


cdef extern from "platform_dep.h":
    void *alloca(int)

cdef class ParticleSmoothOperation:
    # We assume each will allocate and define their own temporary storage
    cdef kernel_func sph_kernel
    cdef public object nvals
    cdef np.float64_t DW[3]
    cdef int nfields
    cdef int maxn
    cdef bint periodicity[3]
    # Note that we are preallocating here, so this is *not* threadsafe.
    cdef void (*pos_setup)(np.float64_t ipos[3], np.float64_t opos[3])
    cdef void neighbor_process(self, int dim[3], np.float64_t left_edge[3],
                               np.float64_t dds[3], np.float64_t[:,:] ppos,
                               np.float64_t **fields,
                               np.int64_t[:] doffs, np.int64_t **nind,
                               np.int64_t[:] pinds, np.int64_t[:] pcounts,
                               np.int64_t offset, np.float64_t **index_fields,
                               OctreeContainer octree, np.int64_t domain_id,
                               int *nsize, np.float64_t[:,:] oct_left_edges,
                               np.float64_t[:,:] oct_dds, DistanceQueue dq)
    cdef int neighbor_search(self, np.float64_t pos[3], OctreeContainer octree,
                             np.int64_t **nind, int *nsize,
                             np.int64_t nneighbors, np.int64_t domain_id,
                             Oct **oct = ?, int extra_layer = ?)
    cdef void neighbor_process_particle(self, np.float64_t cpos[3],
                               np.float64_t[:,:] ppos,
                               np.float64_t **fields,
                               np.int64_t[:] doffs, np.int64_t **nind,
                               np.int64_t[:] pinds, np.int64_t[:] pcounts,
                               np.int64_t offset,
                               np.float64_t **index_fields,
                               OctreeContainer octree, np.int64_t domain_id,
                               int *nsize, DistanceQueue dq)
    cdef void neighbor_find(self,
                            np.int64_t nneighbors,
                            np.int64_t *nind,
                            np.int64_t[:] doffs,
                            np.int64_t[:] pcounts,
                            np.int64_t[:] pinds,
                            np.float64_t[:,:] ppos,
                            np.float64_t cpos[3],
                            np.float64_t[:,:] oct_left_edges,
                            np.float64_t[:,:] oct_dds, DistanceQueue dq)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq)



# distutils: libraries = STD_LIBS
# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG OMP_ARGS
# distutils: extra_link_args = CPP14_FLAG OMP_ARGS
"""
Simple integrators for the radiative transfer equation



"""


import numpy as np

cimport cython
cimport numpy as np
from cython cimport floating
from libc.math cimport copysign, fabs, log2
from libc.stdlib cimport free, malloc

from yt.utilities.lib.fp_utils cimport fclip, i64clip

from yt.utilities.exceptions import YTDomainOverflow

from yt.utilities.lib.vec3_ops cimport L2_norm, cross, dot, subtract

DEF ORDER_MAX=20
DEF INDEX_MAX_64=2097151
DEF XSHIFT=2
DEF YSHIFT=1
DEF ZSHIFT=0

cdef extern from "math.h":
    double exp(double x) nogil
    float expf(float x) nogil
    long double expl(long double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil
    double fmod(double x, double y) nogil
    double fabs(double x) nogil

cdef extern from "platform_dep.h":
    long int lrint(double x) nogil

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t graycode(np.int64_t x):
    return x^(x>>1)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t igraycode(np.int64_t x):
    cdef np.int64_t i, j
    if x == 0:
        return x
    m = <np.int64_t> ceil(log2(x)) + 1
    i, j = x, 1
    while j < m:
        i = i ^ (x>>j)
        j += 1
    return i

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t direction(np.int64_t x, np.int64_t n):
    #assert x < 2**n
    if x == 0:
        return 0
    elif x%2 == 0:
        return tsb(x-1, n)%n
    else:
        return tsb(x, n)%n

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t tsb(np.int64_t x, np.int64_t width):
    #assert x < 2**width
    cdef np.int64_t i = 0
    while x&1 and i <= width:
        x = x >> 1
        i += 1
    return i

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t bitrange(np.int64_t x, np.int64_t width,
                         np.int64_t start, np.int64_t end):
    return x >> (width-end) & ((2**(end-start))-1)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t rrot(np.int64_t x, np.int64_t i, np.int64_t width):
    i = i%width
    x = (x>>i) | (x<<width-i)
    return x&(2**width-1)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t lrot(np.int64_t x, np.int64_t i, np.int64_t width):
    i = i%width
    x = (x<<i) | (x>>width-i)
    return x&(2**width-1)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t transform(np.int64_t entry, np.int64_t direction,
                          np.int64_t width, np.int64_t x):
    return rrot((x^entry), direction + 1, width)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t entry(np.int64_t x):
    if x == 0: return 0
    return graycode(2*((x-1)/2))

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t setbit(np.int64_t x, np.int64_t w, np.int64_t i, np.int64_t b):
    if b == 1:
        return x | 2**(w-i-1)
    elif b == 0:
        return x & ~2**(w-i-1)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def spread_bits(np.uint64_t x):
    return spread_64bits_by2(x)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def compact_bits(np.uint64_t x):
    return compact_64bits_by2(x)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def lsz(np.uint64_t v, int stride = 1, int start = 0):
    cdef int c
    c = start
    while ((np.uint64(1) << np.uint64(c)) & np.uint64(v)):
        c += stride
    return c

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def lsb(np.uint64_t v, int stride = 1, int start = 0):
    cdef int c
    c = start
    while (np.uint64(v) << np.uint64(c)) and not ((np.uint64(1) << np.uint64(c)) & np.uint64(v)):
        c += stride
    return c

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def bitwise_addition(np.uint64_t x, np.int64_t y0,
                     int stride = 1, int start = 0):
    if (y0 == 0): return x
    cdef int end, p, pstart
    cdef list mstr
    cdef np.uint64_t m, y, out
    y = np.uint64(np.abs(y0))
    if (y0 > 0):
        func_ls = lsz
    else:
        func_ls = lsb
    # Continue until all bits added
    p = 0
    out = x
    while (y >> p):
        if (y & (1 << p)):
            # Get end point
            pstart = start + p*stride
            end = func_ls(out,stride=stride,start=pstart)
            # Create mask
            mstr = (end + 1) * ['0']
            for i in range(pstart,end+1,stride):
                mstr[i] = '1'
                m = int(''.join(mstr[::-1]), 2)
            # Invert portion in mask
            # print(mstr[::-1])
            # print(y,p,(pstart,end+1),bin(m),bin(out),bin(~out))
            out = masked_merge_64bit(out, ~out, m)
        # Move to next bit
        p += 1
    return out

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t point_to_hilbert(int order, np.int64_t p[3]):
    cdef np.int64_t h, e, d, l, b, w, i, x
    h = e = d = 0
    for i in range(order):
        l = 0
        for x in range(3):
            b = bitrange(p[3-x-1], order, i, i+1)
            l |= (b<<x)
        l = transform(e, d, 3, l)
        w = igraycode(l)
        e = e ^ lrot(entry(w), d+1, 3)
        d = (d + direction(w, 3) + 1)%3
        h = (h<<3)|w
    return h

#def hilbert_point(dimension, order, h):
#    """
#        Convert an index on the Hilbert curve of the specified dimension and
#        order to a set of point coordinates.
#    """
#    #    The bit widths in this function are:
#    #        p[*]  - order
#    #        h     - order*dimension
#    #        l     - dimension
#    #        e     - dimension
#    hwidth = order*dimension
#    e, d = 0, 0
#    p = [0]*dimension
#    for i in range(order):
#        w = utils.bitrange(h, hwidth, i*dimension, i*dimension+dimension)
#        l = utils.graycode(w)
#        l = itransform(e, d, dimension, l)
#        for j in range(dimension):
#            b = utils.bitrange(l, dimension, j, j+1)
#            p[j] = utils.setbit(p[j], order, i, b)
#        e = e ^ utils.lrot(entry(w), d+1, dimension)
#        d = (d + direction(w, dimension) + 1)%dimension
#    return p

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef void hilbert_to_point(int order, np.int64_t h, np.int64_t *p):
    cdef np.int64_t hwidth, e, d, w, l, b
    cdef int i, j
    hwidth = 3 * order
    e = d = p[0] = p[1] = p[2] = 0
    for i in range(order):
        w = bitrange(h, hwidth, i*3, i*3+3)
        l = graycode(w)
        l = lrot(l, d +1, 3)^e
        for j in range(3):
            b = bitrange(l, 3, j, j+1)
            p[j] = setbit(p[j], order, i, b)
        e = e ^ lrot(entry(w), d+1, 3)
        d = (d + direction(w, 3) + 1)%3

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_hilbert_indices(int order, np.ndarray[np.int64_t, ndim=2] left_index):
    # This is inspired by the scurve package by user cortesi on GH.
    cdef int i
    cdef np.int64_t p[3]
    cdef np.ndarray[np.int64_t, ndim=1] hilbert_indices
    hilbert_indices = np.zeros(left_index.shape[0], 'int64')
    for i in range(left_index.shape[0]):
        p[0] = left_index[i, 0]
        p[1] = left_index[i, 1]
        p[2] = left_index[i, 2]
        hilbert_indices[i] = point_to_hilbert(order, p)
    return hilbert_indices

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_hilbert_points(int order, np.ndarray[np.int64_t, ndim=1] indices):
    # This is inspired by the scurve package by user cortesi on GH.
    cdef int i, j
    cdef np.int64_t p[3]
    cdef np.ndarray[np.int64_t, ndim=2] positions
    positions = np.zeros((indices.shape[0], 3), 'int64')
    for i in range(indices.shape[0]):
        hilbert_to_point(order, indices[i], p)
        for j in range(3):
            positions[i, j] = p[j]
    return positions

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.uint64_t point_to_morton(np.uint64_t p[3]):
    # Weird indent thing going on... also, should this reference the pxd func?
    return encode_morton_64bit(p[0],p[1],p[2])

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef void morton_to_point(np.uint64_t mi, np.uint64_t *p):
    decode_morton_64bit(mi,p)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_index(np.ndarray[np.uint64_t, ndim=1] left_index):
    cdef int j
    cdef np.uint64_t morton_index
    cdef np.uint64_t p[3]
    for j in range(3):
        if left_index[j] >= INDEX_MAX_64:
            raise ValueError("Point exceeds max ({}) ".format(INDEX_MAX_64)+
                             "for 64bit interleave.")
        p[j] = left_index[j]
    morton_index = point_to_morton(p)
    return morton_index

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_indices(np.ndarray[np.uint64_t, ndim=2] left_index):
    cdef np.int64_t i
    cdef int j
    cdef np.ndarray[np.uint64_t, ndim=1] morton_indices
    cdef np.uint64_t p[3]
    morton_indices = np.zeros(left_index.shape[0], 'uint64')
    for i in range(left_index.shape[0]):
        for j in range(3):
            if left_index[i, j] >= INDEX_MAX_64:
                raise ValueError("Point exceeds max ({}) ".format(INDEX_MAX_64)+
                                 "for 64bit interleave.")
            p[j] = left_index[i, j]
        morton_indices[i] = point_to_morton(p)
    return morton_indices

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_indices_unravel(np.ndarray[np.uint64_t, ndim=1] left_x,
                               np.ndarray[np.uint64_t, ndim=1] left_y,
                               np.ndarray[np.uint64_t, ndim=1] left_z):
    cdef np.int64_t i
    cdef np.ndarray[np.uint64_t, ndim=1] morton_indices
    cdef np.uint64_t p[3]
    morton_indices = np.zeros(left_x.shape[0], 'uint64')
    for i in range(left_x.shape[0]):
        p[0] = left_x[i]
        p[1] = left_y[i]
        p[2] = left_z[i]
        for j in range(3):
            if p[j] >= INDEX_MAX_64:
                raise ValueError("Point exceeds max ({}) ".format(INDEX_MAX_64)+
                                 "for 64 bit interleave.")
        morton_indices[i] = point_to_morton(p)
    return morton_indices

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_point(np.uint64_t index):
    cdef int j
    cdef np.uint64_t p[3]
    cdef np.ndarray[np.uint64_t, ndim=1] position
    position = np.zeros(3, 'uint64')
    morton_to_point(index, p)
    for j in range(3):
        position[j] = p[j]
    return position

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_points(np.ndarray[np.uint64_t, ndim=1] indices):
    # This is inspired by the scurve package by user cortesi on GH.
    cdef int i, j
    cdef np.uint64_t p[3]
    cdef np.ndarray[np.uint64_t, ndim=2] positions
    positions = np.zeros((indices.shape[0], 3), 'uint64')
    for i in range(indices.shape[0]):
        morton_to_point(indices[i], p)
        for j in range(3):
            positions[i, j] = p[j]
    return positions

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_neighbors_coarse(mi1, max_index1, periodic, nn):
    cdef int i
    cdef np.uint32_t ntot
    cdef np.ndarray[np.uint32_t, ndim=2] index = np.zeros((2*nn+1,3), dtype='uint32')
    cdef np.ndarray[np.uint64_t, ndim=2] ind1_n = np.zeros((2*nn+1,3), dtype='uint64')
    cdef np.ndarray[np.uint64_t, ndim=1] neighbors = np.zeros((2*nn+1)**3, dtype='uint64')
    cdef bint periodicity[3]
    if periodic:
        for i in range(3): periodicity[i] = 1
    else:
        for i in range(3): periodicity[i] = 0
    ntot = morton_neighbors_coarse(mi1, max_index1, periodicity, nn,
                                   index, ind1_n, neighbors)
    return np.resize(neighbors, (ntot,))

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.uint32_t morton_neighbors_coarse(np.uint64_t mi1, np.uint64_t max_index1,
                                         bint periodicity[3], np.uint32_t nn,
                                         np.uint32_t[:,:] index,
                                         np.uint64_t[:,:] ind1_n,
                                         np.uint64_t[:] neighbors):
    cdef np.uint32_t ntot = 0
    cdef np.uint64_t ind1[3]
    cdef np.uint32_t count[3]
    cdef np.uint32_t origin[3]
    cdef np.int64_t adv
    cdef int i, j, k, ii, ij, ik
    for i in range(3):
        count[i] = 0
        origin[i] = 0
    # Get indices
    decode_morton_64bit(mi1,ind1)
    # Determine which directions are valid
    for j,i in enumerate(range(-nn,(nn+1))):
        if i == 0:
            for k in range(3):
                ind1_n[j,k] = ind1[k]
                index[count[k],k] = j
                origin[k] = count[k]
                count[k] += 1
        else:
            for k in range(3):
                adv = <np.int64_t>((<np.int64_t>ind1[k]) + i)
                if (adv < 0):
                    if periodicity[k]:
                        while adv < 0:
                            adv += max_index1
                        ind1_n[j,k] = <np.uint64_t>(adv % max_index1)
                    else:
                        continue
                elif (adv >= max_index1):
                    if periodicity[k]:
                        ind1_n[j,k] = <np.uint64_t>(adv % max_index1)
                    else:
                        continue
                else:
                    ind1_n[j,k] = <np.uint64_t>(adv)
                # print(i,k,adv,max_index1,ind1_n[j,k],adv % max_index1)
                index[count[k],k] = j
                count[k] += 1
    # Iterate over ever combinations
    for ii in range(count[0]):
        i = index[ii,0]
        for ij in range(count[1]):
            j = index[ij,1]
            for ik in range(count[2]):
                k = index[ik,2]
                if (ii != origin[0]) or (ij != origin[1]) or (ik != origin[2]):
                    neighbors[ntot] = encode_morton_64bit(ind1_n[i,0],
                                                          ind1_n[j,1],
                                                          ind1_n[k,2])
                    ntot += 1
    return ntot

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_neighbors_refined(mi1, mi2, max_index1, max_index2, periodic, nn):
    cdef int i
    cdef np.uint32_t ntot
    cdef np.ndarray[np.uint32_t, ndim=2] index = np.zeros((2*nn+1,3), dtype='uint32')
    cdef np.ndarray[np.uint64_t, ndim=2] ind1_n = np.zeros((2*nn+1,3), dtype='uint64')
    cdef np.ndarray[np.uint64_t, ndim=2] ind2_n = np.zeros((2*nn+1,3), dtype='uint64')
    cdef np.ndarray[np.uint64_t, ndim=1] neighbors1 = np.zeros((2*nn+1)**3, dtype='uint64')
    cdef np.ndarray[np.uint64_t, ndim=1] neighbors2 = np.zeros((2*nn+1)**3, dtype='uint64')
    cdef bint periodicity[3]
    if periodic:
        for i in range(3): periodicity[i] = 1
    else:
        for i in range(3): periodicity[i] = 0
    ntot = morton_neighbors_refined(mi1, mi2, max_index1, max_index2,
                                    periodicity, nn,
                                    index, ind1_n, ind2_n,
                                    neighbors1, neighbors2)
    return np.resize(neighbors1, (ntot,)), np.resize(neighbors2, (ntot,))

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.uint32_t morton_neighbors_refined(np.uint64_t mi1, np.uint64_t mi2,
                                          np.uint64_t max_index1, np.uint64_t max_index2,
                                          bint periodicity[3], np.uint32_t nn,
                                          np.uint32_t[:,:] index,
                                          np.uint64_t[:,:] ind1_n,
                                          np.uint64_t[:,:] ind2_n,
                                          np.uint64_t[:] neighbors1,
                                          np.uint64_t[:] neighbors2):
    cdef np.uint32_t ntot = 0
    cdef np.uint64_t ind1[3]
    cdef np.uint64_t ind2[3]
    cdef np.uint32_t count[3]
    cdef np.uint32_t origin[3]
    cdef np.int64_t adv, maj, rem, adv1
    cdef int i, j, k, ii, ij, ik
    for i in range(3):
        count[i] = 0
        origin[i] = 0
    # Get indices
    decode_morton_64bit(mi1,ind1)
    decode_morton_64bit(mi2,ind2)
    # Determine which directions are valid
    for j,i in enumerate(range(-nn,(nn+1))):
        if i == 0:
            for k in range(3):
                ind1_n[j,k] = ind1[k]
                ind2_n[j,k] = ind2[k]
                index[count[k],k] = j
                origin[k] = count[k]
                count[k] += 1
        else:
            for k in range(3):
                adv = <np.int64_t>(ind2[k] + i)
                maj = adv / (<np.int64_t>max_index2)
                rem = adv % (<np.int64_t>max_index2)
                if adv < 0:
                    adv1 = <np.int64_t>(ind1[k] + (maj-1))
                    if adv1 < 0:
                        if periodicity[k]:
                            while adv1 < 0:
                                adv1 += max_index1
                            ind1_n[j,k] = <np.uint64_t>adv1
                        else:
                            continue
                    else:
                        ind1_n[j,k] = <np.uint64_t>adv1
                    while adv < 0:
                        adv += max_index2
                    ind2_n[j,k] = <np.uint64_t>adv
                elif adv >= max_index2:
                    adv1 = <np.int64_t>(ind1[k] + maj)
                    if adv1 >= max_index1:
                        if periodicity[k]:
                            ind1_n[j,k] = <np.uint64_t>(adv1 % <np.int64_t>max_index1)
                        else:
                            continue
                    else:
                        ind1_n[j,k] = <np.uint64_t>adv1
                    ind2_n[j,k] = <np.uint64_t>rem
                else:
                    ind1_n[j,k] = ind1[k]
                    ind2_n[j,k] = <np.uint64_t>(adv)
                index[count[k],k] = j
                count[k] += 1
    # Iterate over ever combinations
    for ii in range(count[0]):
        i = index[ii,0]
        for ij in range(count[1]):
            j = index[ij,1]
            for ik in range(count[2]):
                k = index[ik,2]
                if (ii != origin[0]) or (ij != origin[1]) or (ik != origin[2]):
                    neighbors1[ntot] = encode_morton_64bit(ind1_n[i,0],
                                                           ind1_n[j,1],
                                                           ind1_n[k,2])
                    neighbors2[ntot] = encode_morton_64bit(ind2_n[i,0],
                                                           ind2_n[j,1],
                                                           ind2_n[k,2])
                    ntot += 1
    return ntot

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_neighbor_periodic(np.ndarray[np.uint64_t,ndim=1] p,
                             list dim_list, list num_list,
                             np.uint64_t max_index):
    cdef np.uint64_t p1[3]
    cdef int j, dim, num
    for j in range(3):
        p1[j] = np.uint64(p[j])
    for dim,num in zip(dim_list,num_list):
        p1[dim] = np.uint64((np.int64(p[dim]) + num) % max_index)
    return np.int64(point_to_morton(p1))

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_neighbor_bounded(np.ndarray[np.uint64_t,ndim=1] p,
                            list dim_list, list num_list,
                            np.uint64_t max_index):
    cdef np.int64_t x
    cdef np.uint64_t p1[3]
    cdef int j, dim, num
    for j in range(3):
        p1[j] = np.uint64(p[j])
    for dim,num in zip(dim_list,num_list):
        x = np.int64(p[dim]) + num
        if (x >= 0) and (x < max_index):
            p1[dim] = np.uint64(x)
        else:
            return np.int64(-1)
    return np.int64(point_to_morton(p1))

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_neighbor(np.ndarray[np.uint64_t,ndim=1] p,
                    list dim_list, list num_list,
                    np.uint64_t max_index, periodic = False):
    if periodic:
        return morton_neighbor_periodic(p, dim_list, num_list, max_index)
    else:
        return morton_neighbor_bounded(p, dim_list, num_list, max_index)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_neighbors(np.ndarray[np.uint64_t,ndim=1] mi,
                         int order = ORDER_MAX, periodic = False):
    """Returns array of neighboring morton indices"""
    # Declare
    cdef int i, j, k, l, n
    cdef np.uint64_t max_index
    cdef np.ndarray[np.uint64_t, ndim=2] p
    cdef np.int64_t nmi
    cdef np.ndarray[np.uint64_t, ndim=1] mi_neighbors
    p = get_morton_points(mi)
    mi_neighbors = np.zeros(26*mi.shape[0], 'uint64')
    n = 0
    max_index = np.int64(1 << order)
    # Define function
    if periodic:
        fneighbor = morton_neighbor_periodic
    else:
        fneighbor = morton_neighbor_bounded
    for i in range(mi.shape[0]):
        for j in range(3):
            # +1 in dimension j
            nmi = fneighbor(p[i,:],[j],[+1],max_index)
            if nmi > 0:
                mi_neighbors[n] = np.uint64(nmi)
                n+=1
                # +/- in dimension k
                for k in range(j+1,3):
                    # +1 in dimension k
                    nmi = fneighbor(p[i,:],[j,k],[+1,+1],max_index)
                    if nmi > 0:
                        mi_neighbors[n] = np.uint64(nmi)
                        n+=1
                        # +/- in dimension l
                        for l in range(k+1,3):
                            nmi = fneighbor(p[i,:],[j,k,l],[+1,+1,+1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                            nmi = fneighbor(p[i,:],[j,k,l],[+1,+1,-1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                    # -1 in dimension k
                    nmi = fneighbor(p[i,:],[j,k],[+1,-1],max_index)
                    if nmi > 0:
                        mi_neighbors[n] = np.uint64(nmi)
                        n+=1
                        # +/- in dimension l
                        for l in range(k+1,3):
                            nmi = fneighbor(p[i,:],[j,k,l],[+1,-1,+1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                            nmi = fneighbor(p[i,:],[j,k,l],[+1,-1,-1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
            # -1 in dimension j
            nmi = fneighbor(p[i,:],[j],[-1],max_index)
            if nmi > 0:
                mi_neighbors[n] = np.uint64(nmi)
                n+=1
                # +/- in dimension k
                for k in range(j+1,3):
                    # +1 in dimension k
                    nmi = fneighbor(p[i,:],[j,k],[-1,+1],max_index)
                    if nmi > 0:
                        mi_neighbors[n] = np.uint64(nmi)
                        n+=1
                        # +/- in dimension l
                        for l in range(k+1,3):
                            nmi = fneighbor(p[i,:],[j,k,l],[-1,+1,+1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                            nmi = fneighbor(p[i,:],[j,k,l],[-1,+1,-1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                    # -1 in dimension k
                    nmi = fneighbor(p[i,:],[j,k],[-1,-1],max_index)
                    if nmi > 0:
                        mi_neighbors[n] = np.uint64(nmi)
                        n+=1
                        # +/- in dimension l
                        for l in range(k+1,3):
                            nmi = fneighbor(p[i,:],[j,k,l],[-1,-1,+1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
                            nmi = fneighbor(p[i,:],[j,k,l],[-1,-1,-1],max_index)
                            if nmi > 0:
                                mi_neighbors[n] = np.uint64(nmi)
                                n+=1
    mi_neighbors = np.resize(mi_neighbors,(n,))
    return np.unique(np.hstack([mi,mi_neighbors]))

def ifrexp_cy(np.float64_t x):
    cdef np.int64_t e, m
    m = ifrexp(x, &e)
    return m,e

def msdb_cy(np.int64_t a, np.int64_t b):
    return msdb(a,b)

def msdb_cy(np.int64_t a, np.int64_t b):
    return msdb(a,b)

def xor_msb_cy(np.float64_t a, np.float64_t b):
    return xor_msb(a,b)

def morton_qsort_swap(np.ndarray[np.uint64_t, ndim=1] ind,
                      np.uint64_t a, np.uint64_t b):
    # http://www.geeksforgeeks.org/iterative-quick-sort/
    cdef np.int64_t t = ind[a]
    ind[a] = ind[b]
    ind[b] = t

def morton_qsort_partition(np.ndarray[floating, ndim=2] pos,
                           np.int64_t l, np.int64_t h,
                           np.ndarray[np.uint64_t, ndim=1] ind,
                           use_loop = False):
    # Initialize
    cdef int k
    cdef np.int64_t i, j
    cdef np.float64_t ppos[3]
    cdef np.float64_t ipos[3]
    cdef np.uint64_t done, pivot
    if use_loop:
        # http://www.geeksforgeeks.org/iterative-quick-sort/
        # A bit slower
        # Set starting point & pivot
        i = (l - 1)
        for k in range(3):
            ppos[k] = pos[ind[h],k]
        # Loop over array moving ind for points smaller than pivot to front
        for j in range(l, h):
            for k in range(3):
                ipos[k] = pos[ind[j],k]
            if compare_floats_morton(ipos,ppos):
                i+=1
                morton_qsort_swap(ind,i,j)
        # Swap the pivot to the midpoint in the partition
        i+=1
        morton_qsort_swap(ind,i,h)
        return i
    else:
        # Set starting point & pivot
        i = l-1
        j = h
        done = 0
        pivot = ind[h]
        for k in range(3):
            ppos[k] = pos[pivot,k]
        # Loop until entire array processed
        while not done:
            # Process bottom
            while not done:
                i+=1
                if i == j:
                    done = 1
                    break
                for k in range(3):
                    ipos[k] = pos[ind[i],k]
                if compare_floats_morton(ppos,ipos):
                    ind[j] = ind[i]
                    break
            # Process top
            while not done:
                j-=1
                if j == i:
                    done = 1
                    break
                for k in range(3):
                    ipos[k] = pos[ind[j],k]
                if compare_floats_morton(ipos,ppos):
                    ind[i] = ind[j]
                    break
        ind[j] = pivot
    return j

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_qsort_recursive(np.ndarray[floating, ndim=2] pos,
                           np.int64_t l, np.int64_t h,
                           np.ndarray[np.uint64_t, ndim=1] ind,
                           use_loop = False):
    # http://www.geeksforgeeks.org/iterative-quick-sort/
    cdef np.int64_t p
    if (l < h):
        p = morton_qsort_partition(pos, l, h, ind, use_loop=use_loop)
        morton_qsort_recursive(pos, l, p-1, ind, use_loop=use_loop)
        morton_qsort_recursive(pos, p+1, h, ind, use_loop=use_loop)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_qsort_iterative(np.ndarray[floating, ndim=2] pos,
                           np.int64_t l, np.int64_t h,
                           np.ndarray[np.uint64_t, ndim=1] ind,
                           use_loop = False):
    # http://www.geeksforgeeks.org/iterative-quick-sort/
    # Auxiliary stack
    cdef np.ndarray[np.int64_t, ndim=1] stack = np.zeros(h-l+1, dtype=np.int64)
    cdef np.int64_t top = -1
    cdef np.int64_t p
    top+=1
    stack[top] = l
    top+=1
    stack[top] = h
    # Pop from stack until it's empty
    while (top >= 0):
        # Get next set
        h = stack[top]
        top-=1
        l = stack[top]
        top-=1
        # Partition
        p = morton_qsort_partition(pos, l, h, ind, use_loop=use_loop)
        # Add left partition to the stack
        if (p-1) > l:
            top+=1
            stack[top] = l
            top+=1
            stack[top] = p - 1
        # Add right partition to the stack
        if (p+1) < h:
            top+=1
            stack[top] = p + 1
            top+=1
            stack[top] = h

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def morton_qsort(np.ndarray[floating, ndim=2] pos,
                 np.int64_t l, np.int64_t h,
                 np.ndarray[np.uint64_t, ndim=1] ind,
                 recursive = False,
                 use_loop = False):
    #get_morton_argsort1(pos,l,h,ind)
    if recursive:
        morton_qsort_recursive(pos,l,h,ind,use_loop=use_loop)
    else:
        morton_qsort_iterative(pos,l,h,ind,use_loop=use_loop)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_morton_argsort1(np.ndarray[floating, ndim=2] pos,
                        np.int64_t start, np.int64_t end,
                        np.ndarray[np.uint64_t, ndim=1] ind):
    # Return if only one position selected
    if start >= end: return
    # Initialize
    cdef np.int64_t top
    top = morton_qsort_partition(pos,start,end,ind)
    # Do remaining parts on either side of pivot, sort side first
    if (top-1-start < end-(top+1)):
        get_morton_argsort1(pos,start,top-1,ind)
        get_morton_argsort1(pos,top+1,end,ind)
    else:
        get_morton_argsort1(pos,top+1,end,ind)
        get_morton_argsort1(pos,start,top-1,ind)
    return

def compare_morton(np.ndarray[floating, ndim=1] p0, np.ndarray[floating, ndim=1] q0):
    cdef np.float64_t p[3]
    cdef np.float64_t q[3]
    # cdef np.int64_t iep,ieq,imp,imq
    cdef int j
    for j in range(3):
        p[j] = p0[j]
        q[j] = q0[j]
        # imp = ifrexp(p[j],&iep)
        # imq = ifrexp(q[j],&ieq)
        # print(j,p[j],q[j],xor_msb(p[j],q[j]),'m=',imp,imq,'e=',iep,ieq)
    return compare_floats_morton(p,q)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t position_to_morton(np.ndarray[floating, ndim=1] pos_x,
                        np.ndarray[floating, ndim=1] pos_y,
                        np.ndarray[floating, ndim=1] pos_z,
                        np.float64_t dds[3], np.float64_t DLE[3],
                        np.float64_t DRE[3],
                        np.ndarray[np.uint64_t, ndim=1] ind,
                        int filter):
    cdef np.uint64_t mi
    cdef np.uint64_t ii[3]
    cdef np.float64_t p[3]
    cdef np.int64_t i, j, use
    cdef np.uint64_t DD[3]
    cdef np.uint64_t FLAG = ~(<np.uint64_t>0)
    for i in range(3):
        DD[i] = <np.uint64_t> ((DRE[i] - DLE[i]) / dds[i])
    for i in range(pos_x.shape[0]):
        use = 1
        p[0] = <np.float64_t> pos_x[i]
        p[1] = <np.float64_t> pos_y[i]
        p[2] = <np.float64_t> pos_z[i]
        for j in range(3):
            if p[j] < DLE[j] or p[j] > DRE[j]:
                if filter == 1:
                    # We only allow 20 levels, so this is inaccessible
                    use = 0
                    break
                return i
            ii[j] = <np.uint64_t> ((p[j] - DLE[j])/dds[j])
            ii[j] = i64clip(ii[j], 0, DD[j] - 1)
        if use == 0:
            ind[i] = FLAG
            continue
        ind[i] = encode_morton_64bit(ii[0],ii[1],ii[2])
    return pos_x.shape[0]

def compute_morton(np.ndarray pos_x, np.ndarray pos_y, np.ndarray pos_z,
                   domain_left_edge, domain_right_edge, filter_bbox = False,
                   order = ORDER_MAX):
    cdef int i
    cdef int filter
    if filter_bbox:
        filter = 1
    else:
        filter = 0
    cdef np.float64_t dds[3]
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]
    for i in range(3):
        DLE[i] = domain_left_edge[i]
        DRE[i] = domain_right_edge[i]
        dds[i] = (DRE[i] - DLE[i]) / (1 << order)
    cdef np.ndarray[np.uint64_t, ndim=1] ind
    ind = np.zeros(pos_x.shape[0], dtype="uint64")
    cdef np.int64_t rv
    if pos_x.dtype == np.float32:
        rv = position_to_morton[np.float32_t](
                pos_x, pos_y, pos_z, dds, DLE, DRE, ind,
                filter)
    elif pos_x.dtype == np.float64:
        rv = position_to_morton[np.float64_t](
                pos_x, pos_y, pos_z, dds, DLE, DRE, ind,
                filter)
    else:
        print("Could not identify dtype.", pos_x.dtype)
        raise NotImplementedError
    if rv < pos_x.shape[0]:
        mis = (pos_x.min(), pos_y.min(), pos_z.min())
        mas = (pos_x.max(), pos_y.max(), pos_z.max())
        raise YTDomainOverflow(mis, mas,
                               domain_left_edge, domain_right_edge)
    return ind

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def dist(np.ndarray[np.float64_t, ndim=1] p0, np.ndarray[np.float64_t, ndim=1] q0):
    cdef int j
    cdef np.float64_t p[3]
    cdef np.float64_t q[3]
    for j in range(3):
        p[j] = p0[j]
        q[j] = q0[j]
    return euclidean_distance(p,q)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def dist_to_box(np.ndarray[np.float64_t, ndim=1] p,
                np.ndarray[np.float64_t, ndim=1] cbox,
                np.float64_t rbox):
    cdef int j
    cdef np.float64_t d = 0.0
    for j in range(3):
        d+= max((cbox[j]-rbox)-p[j],0.0,p[j]-(cbox[j]+rbox))**2
    return np.sqrt(d)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def solution_radius(np.ndarray[np.float64_t, ndim=2] P, int k, np.uint64_t i,
                    np.ndarray[np.uint64_t, ndim=1] idx, int order,
                    np.ndarray[np.float64_t, ndim=1] DLE,
                    np.ndarray[np.float64_t, ndim=1] DRE):
    c = np.zeros(3, dtype=np.float64)
    return quadtree_box(P[i,:],P[idx[k-1],:],order,DLE,DRE,c)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def knn_direct(np.ndarray[np.float64_t, ndim=2] P, np.uint64_t k, np.uint64_t i,
               np.ndarray[np.uint64_t, ndim=1] idx, return_dist = False,
               return_rad = False):
    """Directly compute the k nearest neighbors by sorting on distance.

    Args:
        P (np.ndarray): (N,d) array of points to search sorted by Morton order.
        k (int): number of nearest neighbors to find.
        i (int): index of point that nearest neighbors should be found for.
        idx (np.ndarray): indices of points from P to be considered.
        return_dist (Optional[bool]): If True, distances to the k nearest
            neighbors are also returned (in order of proximity).
            (default = False)
        return_rad (Optional[bool]): If True, distance to farthest nearest
            neighbor is also returned. This is set to False if return_dist is
            True. (default = False)

    Returns:
        np.ndarray: Indices of k nearest neighbors to point i.

    """
    cdef int j,m
    cdef np.int64_t[:] sort_fwd
    cdef np.float64_t[:] ipos
    cdef np.float64_t[:] jpos
    cdef np.float64_t[:] dist = np.zeros(len(idx), dtype='float64')
    ipos = np.zeros(3)
    jpos = np.zeros(3)
    for m in range(3):
        ipos[m] = P[i,m]
    for j in range(len(idx)):
        for m in range(3):
            jpos[m] = P[idx[j],m]
        dist[j] = euclidean_distance(ipos, jpos)
    # casting to uint64 for compatibility with 32 bits systems
    # see https://github.com/yt-project/yt/issues/3656
    sort_fwd = np.argsort(dist, kind='mergesort')[:k].astype(np.int64, copy=False)
    if return_dist:
        return np.array(idx)[sort_fwd], np.array(dist)[sort_fwd]
    elif return_rad:
        return np.array(idx)[sort_fwd], np.array(dist)[sort_fwd][k-1]
    else:
        return np.array(idx)[sort_fwd]

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def quadtree_box(np.ndarray[np.float64_t, ndim=1] p,
                 np.ndarray[np.float64_t, ndim=1] q, int order,
                 np.ndarray[np.float64_t, ndim=1] DLE,
                 np.ndarray[np.float64_t, ndim=1] DRE,
                 np.ndarray[np.float64_t, ndim=1] c):
    # Declare & transfer values to ctypes
    cdef int j
    cdef np.float64_t ppos[3]
    cdef np.float64_t qpos[3]
    cdef np.float64_t rbox
    cdef np.float64_t cbox[3]
    cdef np.float64_t DLE1[3]
    cdef np.float64_t DRE1[3]
    for j in range(3):
        ppos[j] = p[j]
        qpos[j] = q[j]
        DLE1[j] = DLE[j]
        DRE1[j] = DRE[j]
    # Get smallest box containing p & q
    rbox = smallest_quadtree_box(ppos,qpos,order,DLE1,DRE1,
                                 &cbox[0],&cbox[1],&cbox[2])
    # Transfer values to python array
    for j in range(3):
        c[j] = cbox[j]
    return rbox


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def csearch_morton(np.ndarray[np.float64_t, ndim=2] P, int k, np.uint64_t i,
                   np.ndarray[np.uint64_t, ndim=1] Ai,
                   np.uint64_t l, np.uint64_t h, int order,
                   np.ndarray[np.float64_t, ndim=1] DLE,
                   np.ndarray[np.float64_t, ndim=1] DRE, int nu = 4):
    """Expand search concentrically to determine set of k nearest neighbors for
    point i.

    Args:
        P (np.ndarray): (N,d) array of points to search sorted by Morton order.
        k (int): number of nearest neighbors to find.
        i (int): index of point that nearest neighbors should be found for.
        Ai (np.ndarray): (N,k) array of partial nearest neighbor indices.
        l (int): index of lowest point to consider in addition to Ai.
        h (int): index of highest point to consider in addition to Ai.
        order (int): Maximum depth that Morton order quadtree should reach.
        DLE (np.float64[3]): 3 floats defining domain lower bounds in each dim.
        DRE (np.float64[3]): 3 floats defining domain upper bounds in each dim.
        nu (int): minimum number of points before a direct knn search is
            performed. (default = 4)

    Returns:
        np.ndarray: (N,k) array of nearest neighbor indices.

    Raises:
        ValueError: If l<i<h. l and h must be on the same side of i.

    """
    cdef int j
    cdef np.uint64_t m
    # Make sure that h and l are both larger/smaller than i
    if (l < i) and (h > i):
        raise ValueError("Both l and h must be on the same side of i.")
    m = np.uint64((h + l)/2)
    # New range is small enough to consider directly
    if (h-l) < nu:
        if m > i:
            return knn_direct(P,k,i,np.hstack((Ai,np.arange(l,h+1,dtype=np.uint64))))
        else:
            return knn_direct(P,k,i,np.hstack((np.arange(l,h+1,dtype=np.uint64),Ai)))
    # Add middle point
    if m > i:
        Ai,rad_Ai = knn_direct(P,k,i,np.hstack((Ai,m)).astype(np.uint64),return_rad=True)
    else:
        Ai,rad_Ai = knn_direct(P,k,i,np.hstack((m,Ai)).astype(np.uint64),return_rad=True)
    cbox_sol = np.zeros(3,dtype=np.float64)
    rbox_sol = quadtree_box(P[i,:],P[Ai[k-1],:],order,DLE,DRE,cbox_sol)
    # Return current solution if hl box is outside current solution's box
    # Uses actual box
    cbox_hl = np.zeros(3,dtype=np.float64)
    rbox_hl = quadtree_box(P[l,:],P[h,:],order,DLE,DRE,cbox_hl)
    if dist_to_box(cbox_sol,cbox_hl,rbox_hl) >= 1.5*rbox_sol:
        print('{} outside: rad = {}, rbox = {}, dist = {}'.format(m,rad_Ai,rbox_sol,dist_to_box(P[i,:],cbox_hl,rbox_hl)))
        return Ai
    # Expand search to lower/higher indices as needed
    if i < m: # They are already sorted...
        Ai = csearch_morton(P,k,i,Ai,l,m-1,order,DLE,DRE,nu=nu)
        if compare_morton(P[m,:],P[i,:]+dist(P[i,:],P[Ai[k-1],:])):
            Ai = csearch_morton(P,k,i,Ai,m+1,h,order,DLE,DRE,nu=nu)
    else:
        Ai = csearch_morton(P,k,i,Ai,m+1,h,order,DLE,DRE,nu=nu)
        if compare_morton(P[i,:]-dist(P[i,:],P[Ai[k-1],:]),P[m,:]):
            Ai = csearch_morton(P,k,i,Ai,l,m-1,order,DLE,DRE,nu=nu)
    return Ai


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def knn_morton(np.ndarray[np.float64_t, ndim=2] P0, int k, np.uint64_t i0,
               float c = 1.0, int nu = 4, issorted = False, int order = ORDER_MAX,
               np.ndarray[np.float64_t, ndim=1] DLE = np.zeros(3,dtype=np.float64),
               np.ndarray[np.float64_t, ndim=1] DRE = np.zeros(3,dtype=np.float64)):
    """Get the indices of the k nearest neighbors to point i.

    Args:
        P (np.ndarray): (N,d) array of points to search.
        k (int): number of nearest neighbors to find for each point in P.
        i (np.uint64): index of point to find neighbors for.
        c (float): factor determining how many indices before/after i are used
            in the initial search (i-c*k to i+c*k, default = 1.0)
        nu (int): minimum number of points before a direct knn search is
            performed. (default = 4)
        issorted (Optional[bool]): if True, P is assumed to be sorted already
            according to Morton order.
        order (int): Maximum depth that Morton order quadtree should reach.
            If not provided, ORDER_MAX is used.
        DLE (np.ndarray): (d,) array of domain lower bounds in each dimension.
            If not provided, this is determined from the points.
        DRE (np.ndarray): (d,) array of domain upper bounds in each dimension.
            If not provided, this is determined from the points.

    Returns:
        np.ndarray: (N,k) indices of k nearest neighbors for each point in P.
"""
    cdef int j
    cdef np.uint64_t i
    cdef np.int64_t N = P0.shape[0]
    cdef np.ndarray[np.float64_t, ndim=2] P
    cdef np.ndarray[np.uint64_t, ndim=1] sort_fwd = np.arange(N,dtype=np.uint64)
    cdef np.ndarray[np.uint64_t, ndim=1] sort_rev = np.arange(N,dtype=np.uint64)
    cdef np.ndarray[np.uint64_t, ndim=1] Ai
    cdef np.int64_t idxmin, idxmax, u, l, I
    # Sort if necessary
    if issorted:
        P = P0
        i = i0
    else:
        morton_qsort(P0,0,N-1,sort_fwd)
        sort_rev = np.argsort(sort_fwd).astype(np.uint64)
        P = P0[sort_fwd,:]
        i = sort_rev[i0]
    # Check domain and set if singular
    for j in range(3):
        if DLE[j] == DRE[j]:
            DLE[j] = min(P[:,j])
            DRE[j] = max(P[:,j])
    # Get initial guess bassed on position in z-order
    idxmin = <np.int64_t>max(i-c*k, 0)
    idxmax = <np.int64_t>min(i+c*k, N-1)
    Ai = np.hstack((np.arange(idxmin,i,dtype=np.uint64),
                    np.arange(i+1,idxmax+1,dtype=np.uint64)))
    Ai,rad_Ai = knn_direct(P,k,i,Ai,return_rad=True)
    # Get radius of solution
    cbox_Ai = np.zeros(3,dtype=np.float64)
    rbox_Ai = quadtree_box(P[i,:],P[Ai[k-1],:],order,DLE,DRE,cbox_Ai)
    rad_Ai = rbox_Ai
    # Extend upper bound to match lower bound
    if idxmax < (N-1):
        if compare_morton(P[i,:]+rad_Ai,P[idxmax,:]):
            u = i
        else:
            I = 1
            while (idxmax+(2**I) < N) and compare_morton(P[idxmax+(2**I),:],P[i,:]+rad_Ai):
                I+=1
            u = min(idxmax+(2**I),N-1)
            Ai = csearch_morton(P,k,i,Ai,min(idxmax+1,N-1),u,order,DLE,DRE,nu=nu)
    else:
        u = idxmax
    # Extend lower bound to match upper bound
    if idxmin > 0:
        if compare_morton(P[idxmin,:],P[i,:]-rad_Ai):
            l = i
        else:
            I = 1
            while (idxmin-(2**I) >= 0) and compare_morton(P[i,:]-rad_Ai,P[idxmin-(2**I),:]):
                I+=1
            l = max(idxmin-(2**I),0)
            Ai = csearch_morton(P,k,i,Ai,l,max(idxmin-1,0),order,DLE,DRE,nu=nu)
    else:
        l = idxmin
    # Return indices of neighbors in the correct order
    if issorted:
        return Ai
    else:
        return sort_fwd[Ai]

cdef struct PointSet
cdef struct PointSet:
    int count
    # First index is point index, second is xyz
    np.float64_t points[2][3]
    PointSet *next

cdef inline void get_intersection(np.float64_t p0[3], np.float64_t p1[3],
                                  int ax, np.float64_t coord, PointSet *p):
    cdef np.float64_t vec[3]
    cdef np.float64_t t
    for j in range(3):
        vec[j] = p1[j] - p0[j]
    if vec[ax] == 0.0:
        return  # bail if the line is in the plane
    t = (coord - p0[ax])/vec[ax]
    # We know that if they're on opposite sides, it has to intersect.  And we
    # won't get called otherwise.
    for j in range(3):
        p.points[p.count][j] = p0[j] + vec[j] * t
    p.count += 1

@cython.cdivision(True)
def triangle_plane_intersect(int ax, np.float64_t coord,
                             np.ndarray[np.float64_t, ndim=3] triangles):
    cdef np.float64_t p0[3]
    cdef np.float64_t p1[3]
    cdef np.float64_t p2[3]
    cdef np.float64_t E0[3]
    cdef np.float64_t E1[3]
    cdef np.float64_t tri_norm[3]
    cdef np.float64_t plane_norm[3]
    cdef np.float64_t dp
    cdef int i, j, k, count, ntri, nlines
    nlines = 0
    ntri = triangles.shape[0]
    cdef PointSet *first
    cdef PointSet *last
    cdef PointSet *points
    first = last = points = NULL
    for i in range(ntri):
        count = 0

        # Here p0 holds the triangle's zeroth node coordinates,
        # p1 holds the first node's coordinates, and
        # p2 holds the second node's coordinates
        for j in range(3):
            p0[j] = triangles[i, 0, j]
            p1[j] = triangles[i, 1, j]
            p2[j] = triangles[i, 2, j]
            plane_norm[j] = 0.0
        plane_norm[ax] = 1.0
        subtract(p1, p0, E0)
        subtract(p2, p0, E1)
        cross(E0, E1, tri_norm)
        dp = dot(tri_norm, plane_norm)
        dp /= L2_norm(tri_norm)
        # Skip if triangle is close to being parallel to plane.
        if (fabs(dp) > 0.995):
            continue

        # Now for each line segment (01, 12, 20) we check to see how many cross
        # the coordinate of the slice.
        # Here, the components of p2 are either +1 or -1 depending on whether the
        # node's coordinate corresponding to the slice axis is greater than the
        # coordinate of the slice. p2[0] -> node 0; p2[1] -> node 1; p2[2] -> node2
        for j in range(3):
            # Add 0 so that any -0s become +0s. Necessary for consistent determination
            # of plane intersection
            p2[j] = copysign(1.0, triangles[i, j, ax] - coord + 0)
        if p2[0] * p2[1] < 0: count += 1
        if p2[1] * p2[2] < 0: count += 1
        if p2[2] * p2[0] < 0: count += 1
        if count == 2:
            nlines += 1
        elif count == 3:
            raise RuntimeError("It should be geometrically impossible for a plane to"
                               "to intersect all three legs of a triangle. Please contact"
                               "yt developers with your mesh")
        else:
            continue
        points = <PointSet *> malloc(sizeof(PointSet))
        points.count = 0
        points.next = NULL

        # Here p0 and p1 again hold node coordinates
        if p2[0] * p2[1] < 0:
            # intersection of 01 triangle segment with plane
            for j in range(3):
                p0[j] = triangles[i, 0, j]
                p1[j] = triangles[i, 1, j]
            get_intersection(p0, p1, ax, coord, points)
        if p2[1] * p2[2] < 0:
            # intersection of 12 triangle segment with plane
            for j in range(3):
                p0[j] = triangles[i, 1, j]
                p1[j] = triangles[i, 2, j]
            get_intersection(p0, p1, ax, coord, points)
        if p2[2] * p2[0] < 0:
            # intersection of 20 triangle segment with plane
            for j in range(3):
                p0[j] = triangles[i, 2, j]
                p1[j] = triangles[i, 0, j]
            get_intersection(p0, p1, ax, coord, points)
        if last != NULL:
            last.next = points
        if first == NULL:
            first = points
        last = points

    points = first
    cdef np.ndarray[np.float64_t, ndim=3] line_segments
    line_segments = np.empty((nlines, 2, 3), dtype="float64")
    k = 0
    while points != NULL:
        for j in range(3):
            line_segments[k, 0, j] = points.points[0][j]
            line_segments[k, 1, j] = points.points[1][j]
        k += 1
        last = points
        points = points.next
        free(last)
    return line_segments



import pytest
from numpy.testing import assert_raises

from yt.visualization._commons import (
    _swap_arg_pair_order,
    _swap_axes_extents,
    validate_image_name,
)


@pytest.mark.parametrize(
    "name, expected",
    [
        ("noext", "noext.png"),
        ("nothing.png", "nothing.png"),
        ("nothing.pdf", "nothing.pdf"),
        ("version.1.2.3", "version.1.2.3.png"),
    ],
)
def test_default(name, expected):
    result = validate_image_name(name)
    assert result == expected


@pytest.mark.parametrize(
    "name, suffix, expected",
    [
        ("noext", ".png", "noext.png"),
        ("noext", None, "noext.png"),
        ("nothing.png", ".png", "nothing.png"),
        ("nothing.png", None, "nothing.png"),
        ("nothing.png", ".pdf", "nothing.pdf"),
        ("nothing.pdf", ".pdf", "nothing.pdf"),
        ("nothing.pdf", None, "nothing.pdf"),
        ("nothing.pdf", ".png", "nothing.png"),
        ("version.1.2.3", ".png", "version.1.2.3.png"),
        ("version.1.2.3", None, "version.1.2.3.png"),
        ("version.1.2.3", ".pdf", "version.1.2.3.pdf"),
    ],
)
@pytest.mark.filterwarnings(
    r"ignore:Received two valid image formats '\w+' \(from filename\) "
    r"and '\w+' \(from suffix\). The former is ignored.:UserWarning"
)
def test_custom_valid_ext(name, suffix, expected):
    result1 = validate_image_name(name, suffix=suffix)
    assert result1 == expected
    if suffix is not None:
        alt_suffix = suffix.replace(".", "")
        result2 = validate_image_name(name, suffix=alt_suffix)
        assert result2 == expected


def test_extent_swap():
    input_extent = [1, 2, 3, 4]
    expected = [3, 4, 1, 2]
    assert _swap_axes_extents(input_extent) == expected
    assert _swap_axes_extents(tuple(input_extent)) == tuple(expected)


def test_swap_arg_pair_order():
    assert _swap_arg_pair_order(1, 2) == (2, 1)
    assert _swap_arg_pair_order(1, 2, 3, 4, 5, 6) == (2, 1, 4, 3, 6, 5)
    assert_raises(TypeError, _swap_arg_pair_order, 1)
    assert_raises(TypeError, _swap_arg_pair_order, 1, 2, 3)
